from focs._species import *
from species.species_macros.detection import ULTIMATE_DETECTION
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import OCEAN_STANDARD_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
    HAS_INFLUENCE_FOCUS,
    HAS_RESEARCH_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import VERY_BAD_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import AVERAGE_POPULATION
from species.species_macros.research import AVERAGE_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import AVERAGE_STOCKPILE
from species.species_macros.supply import BAD_SUPPLY
from species.species_macros.troops import AVERAGE_DEFENSE_TROOPS

Species(
    name="SP_OURBOOLS",
    description="SP_OURBOOLS_DESC",
    gameplay_description="SP_OURBOOLS_GAMEPLAY_DESC",
    native=True,
    tags=[
        "ORGANIC",
        "ARTISTIC",
        "BAD_INDUSTRY",
        "VERY_BAD_INFLUENCE",
        "BAD_SUPPLY",
        "ULTIMATE_DETECTION",
        "PEDIA_ORGANIC_SPECIES_CLASS",
        "PEDIA_ARTISTIC",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        HAS_RESEARCH_FOCUS,
        HAS_INFLUENCE_FOCUS,
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_RESEARCH",
    likes=[
        "ELERIUM_SPECIAL",
        "FOCUS_RESEARCH",
        "FORTRESS_SPECIAL",
        "FRACTAL_GEODES_SPECIAL",
        "MIMETIC_ALLOY_SPECIAL",
        "PLC_ENVIRONMENTALISM",
        "PLC_ISOLATION",
        "PLC_NO_SUPPLY",
        "RESONANT_MOON_SPECIAL",
    ],
    dislikes=[
        "BLD_BIOTERROR_PROJECTOR",
        # "BLD_CLONING_CENTER"       // disabled content,
        "BLD_CULTURE_ARCHIVES",
        "BLD_GATEWAY_VOID",
        "BLD_IMPERIAL_PALACE",
        "BLD_INTERSPECIES_ACADEMY",
        "BLD_MEGALITH",
        "BLD_STARGATE",
        "KRAKEN_NEST_SPECIAL",
        "PANOPTICON_SPECIAL",
        "PLC_BUREAUCRACY",
        "PLC_INDUSTRIALISM",
        "PLC_SYSTEM_INFRA",
        "TEMPORAL_ANOMALY_SPECIAL",
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        *AVERAGE_RESEARCH,
        *VERY_BAD_INFLUENCE,
        *AVERAGE_STOCKPILE,
        *AVERAGE_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_OURBOOLS"),
        *BAD_SUPPLY,
        *AVERAGE_DEFENSE_TROOPS,
        *ULTIMATE_DETECTION,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
    ],
    environments=OCEAN_STANDARD_EP,
    graphic="icons/species/ourbools.png",
)



import freeOrionAIInterface as fo
from collections.abc import Sequence

from AIDependencies import INVALID_ID
from common.fo_typing import SystemId
from freeorion_tools.caching import cache_for_current_turn
from PlanetUtilsAI import get_capital_sys_id


def _min_max(a, b) -> Sequence:
    """
    Return args in sorted order.

    This function is designed to improve caches performaces,
    when order of arguments doesn't matter we can reduce number of cached calls by
    setting them in the sorted order.

    >>> _min_max(1, 2)
    (1, 2)

    >>> _min_max(2, 1)
    (1, 2)
    """
    return (a, b) if a <= b else (b, a)


def systems_connected(system1: SystemId, system2: SystemId) -> bool:
    """
    Return True if systems connected.
    """
    return _systems_connected(*_min_max(system1, system2))


@cache_for_current_turn
def _systems_connected(system1: SystemId, system2: SystemId) -> bool:
    if system1 == INVALID_ID:
        return False

    # optimization: We cache a set of systems connected to the capital
    # system in a single DFS and check if the systems are both in that set.
    # This allows us to avoid a BFS for each pair of systems.
    # Only in case neither of the system connects to the capital, we need to
    # run the BFS.
    connected_system_set = systems_connected_to_system(get_capital_sys_id())
    sys1_connected_to_capital = system1 in connected_system_set
    sys2_connected_to_capital = system2 in connected_system_set
    if sys1_connected_to_capital and sys2_connected_to_capital:
        # both connected to the capital - so must be connected to each other
        return True

    if sys1_connected_to_capital != sys2_connected_to_capital:
        # Only one connected to the capital - can't be connected to each other
        return False

    # both are not connected to the home system - they may or may not be connected to each other
    return fo.getUniverse().systemsConnected(system1, system2, fo.empireID())


def get_shortest_distance(system_1: SystemId, system_2: SystemId) -> float:
    """
    Return the distance between the systems where objects are located.
    """
    return _get_shortest_distance(*_min_max(system_1, system_2))


@cache_for_current_turn
def _get_shortest_distance(system_1: SystemId, system_2: SystemId) -> float:
    return fo.getUniverse().shortestPathDistance(system_1, system_2)


@cache_for_current_turn
def get_neighbors(sid: SystemId) -> set[SystemId]:
    return set(fo.getUniverse().getImmediateNeighbors(sid, fo.empireID()))


@cache_for_current_turn
def systems_connected_to_system(system_id: SystemId) -> set[SystemId]:
    """
    Use depth-first-search to find connected systems to system_id
    """
    connected_systems = set()
    if system_id == INVALID_ID:
        return connected_systems

    to_visit = [system_id]
    while to_visit:
        current_system = to_visit.pop()
        if current_system in connected_systems:
            continue

        connected_systems.add(current_system)
        to_visit.extend(get_neighbors(current_system))

    return connected_systems


@cache_for_current_turn
def within_n_jumps(system_id: SystemId, n: int) -> frozenset[SystemId]:
    """
    Returns a frozenset of all systems within n jumps from the given system.
    """
    if n < 1:
        return frozenset({system_id})
    elif n == 1:
        return frozenset({system_id} | get_neighbors(system_id))
    tier_minus_2 = within_n_jumps(system_id, n - 2)
    tier_minus_1 = within_n_jumps(system_id, n - 1)
    result = set(tier_minus_1)
    for sys_id in tier_minus_1 - tier_minus_2:
        result.update(get_neighbors(sys_id))
    return frozenset(result)



import freeOrionAIInterface as fo
from collections.abc import Iterable, Mapping, Sequence
from enum import Enum
from logging import debug, error
from typing import (
    NamedTuple,
    Optional,
    Union,
)

from AIDependencies import INVALID_ID, STABILITY_PER_LIKED_FOCUS
from aistate_interface import get_aistate
from common.fo_typing import PlanetId, SpeciesName, SystemId
from empire.colony_builders import get_colony_builders, get_extra_colony_builders
from empire.ship_builders import get_ship_builders
from EnumsAI import FocusType
from freeorion_tools import get_game_rule_int, get_named_real, ppstring
from freeorion_tools.caching import cache_for_current_turn


def sys_name_ids(sys_ids: Iterable[int]) -> str:
    """
    Get a string representation of a list with system_ids.

    The returned string is of the form "[S_id<name>, ...]"

    :return: string representation of the systems in the list
    """
    universe = fo.getUniverse()
    return ppstring([str(universe.getSystem(sys_id)) for sys_id in sys_ids])


def planet_string(planet_ids: Union[PlanetId, Iterable[PlanetId]]) -> str:
    """
    Get a string representation of the passed planets.
    """

    def _safe_planet_name(planet_id):
        planet = fo.getUniverse().getPlanet(planet_id)
        return fo.to_str("P", planet_id, (planet and planet.name) or "?")

    if isinstance(planet_ids, int):
        return _safe_planet_name(planet_ids)
    return ppstring([_safe_planet_name(pid) for pid in planet_ids])


@cache_for_current_turn
def get_capital() -> PlanetId:  # noqa: C901
    """
    Return current empire capital id.

    If no current capital returns planet with biggest population in first not empty group.
    First check all planets with coloniser species, after that with ship builders and at last all inhabited planets.
    """
    universe = fo.getUniverse()
    empire = fo.getEmpire()
    empire_id = empire.empireID
    capital_id = empire.capitalID
    homeworld = universe.getPlanet(capital_id)
    if homeworld:
        if homeworld.owner == empire_id:
            return capital_id
        else:
            debug(
                "Nominal Capitol %s does not appear to be owned by empire %d %s"
                % (homeworld.name, empire_id, empire.name)
            )
    empire_owned_planet_ids = get_owned_planets_by_empire()
    peopled_planets = get_populated_planet_ids(empire_owned_planet_ids)
    if not peopled_planets:
        if empire_owned_planet_ids:
            return empire_owned_planet_ids[0]
        else:
            return INVALID_ID
    try:
        for spec_list in [get_colony_builders(), get_ship_builders(), None]:
            population_id_pairs = []
            for planet_id in peopled_planets:
                planet = universe.getPlanet(planet_id)
                if spec_list is None or planet.speciesName in spec_list:
                    population_id_pairs.append((planet.initialMeterValue(fo.meterType.population), planet_id))
            if population_id_pairs:
                return max(population_id_pairs)[-1]
    except Exception as e:
        error(e, exc_info=True)
    return INVALID_ID  # shouldn't ever reach here


def get_capital_sys_id() -> SystemId:
    """
    Return system id with empire capital.
    :return: system id
    """
    cap_id = get_capital()
    if cap_id == INVALID_ID:
        return INVALID_ID
    else:
        return fo.getUniverse().getPlanet(cap_id).systemID


def get_planets_in__systems_ids(system_ids: Iterable[SystemId]) -> list[PlanetId]:
    """
    Return list of planet ids for system ids list.
    """
    universe = fo.getUniverse()
    planet_ids = set()
    for system_id in system_ids:
        system = universe.getSystem(system_id)
        if system is not None:
            planet_ids.update(system.planetIDs)
    return list(planet_ids)


@cache_for_current_turn
def get_owned_planets_by_empire() -> list[PlanetId]:
    """
    Return list of all planets owned by empire.
    """
    universe = fo.getUniverse()
    empire_id = fo.getEmpire().empireID
    result = []
    for pid in universe.planetIDs:
        planet = universe.getPlanet(pid)
        # even if our universe says we own it, if we can't see it we must have lost it
        if (
            planet
            and not planet.unowned
            and planet.ownedBy(empire_id)
            and universe.getVisibility(pid, empire_id) >= fo.visibility.partial
        ):
            result.append(pid)
    return result


def get_all_owned_planet_ids(planet_ids: Sequence[PlanetId]) -> Sequence[PlanetId]:
    """
    Return list of all owned and populated planet_ids.
    """
    # TODO: remove after refactoring in invasionAI
    # this function result used only to filter out unpopulated planets,
    # after some changes in invasionAI it will became obsolete
    universe = fo.getUniverse()
    result = []
    for pid in planet_ids:
        planet = universe.getPlanet(pid)
        if planet:
            population = planet.initialMeterValue(fo.meterType.population)
            if not planet.unowned or population > 0:
                result.append(pid)
    return result


def get_populated_planet_ids(planet_ids: Sequence[PlanetId]) -> Sequence[PlanetId]:
    """
    Filter planets with population.
    :param planet_ids: list of planets ids
    :return: list of planets ids
    """
    universe = fo.getUniverse()
    return [pid for pid in planet_ids if universe.getPlanet(pid).initialMeterValue(fo.meterType.population) > 0]


@cache_for_current_turn
def get_empire_populated_planets() -> tuple[fo.planet]:
    universe = fo.getUniverse()
    return tuple(universe.getPlanet(pid) for pid in get_populated_planet_ids(get_owned_planets_by_empire()))


def get_systems(planet_ids: Sequence[PlanetId]) -> Sequence[SystemId]:
    """
    Return list of systems containing planet_ids.
    """
    # TODO discuss change return type to set
    universe = fo.getUniverse()
    return [universe.getPlanet(pid).systemID for pid in planet_ids]


class Opinion(NamedTuple):
    likes: set[PlanetId]
    neutral: set[PlanetId]
    dislikes: set[PlanetId]

    def value(self, pid: PlanetId, like_value: float, neutral_value: float, dislike_value: float) -> float:
        """Returns like_value if pid is in likes, dislike_value if pid is in dislikes, else neutral_value"""
        if pid in self.likes:
            return like_value
        elif pid in self.dislikes:
            return dislike_value
        return neutral_value


def get_planet_opinion(feature: Union[str, Enum]) -> Opinion:
    """
    Returns sets of empire planets that like, are neutral and dislike the given feature
    """
    # default: feature not in any like or dislike set, all neutral
    if isinstance(feature, Enum):
        feature = feature.value()
    default = Opinion(set(), set(get_owned_planets_by_empire()), set())
    return _calculate_get_planet_opinions().get(feature, default)


def _get_species_from_colony_building(name: str) -> Optional[SpeciesName]:
    """Extract a species if name is the name of a colony building"""
    building_prefix = "BLD_COL_"
    species_prefix = "SP_"
    if name.startswith(building_prefix):
        return species_prefix + name[len(building_prefix) :]


@cache_for_current_turn
def _planned_species() -> Mapping[PlanetId, SpeciesName]:
    universe = fo.getUniverse()
    production_queue = fo.getEmpire().productionQueue
    planned_species = {}
    colonisation_plans = get_aistate().colonisablePlanetIDs
    for element in production_queue:
        species_name = _get_species_from_colony_building(element.name)
        if species_name:
            planned_species[element.locationID] = species_name
    for pid in get_owned_planets_by_empire():
        planet = universe.getPlanet(pid)
        if planet.speciesName:
            # skip already populated planets
            continue
        # Finished colony buildings are normal buildings for one turn before turning the outpost into a colony
        for building in map(universe.getBuilding, planet.buildingIDs):
            species_name = _get_species_from_colony_building(building.name)
            if species_name:
                planned_species[pid] = species_name
                break
        if pid not in planned_species:
            # Without checking the colonisation plans, the AI may start building a colony and buildings
            # the future species wouldn't like in the same turn.
            plan = colonisation_plans.get(pid)
            if plan:
                planned_species[pid] = plan[1]
    debug(f"Planned species: {planned_species}")
    return planned_species


def _planet_species(pid: PlanetId) -> Optional[fo.species]:
    universe = fo.getUniverse()
    planet = universe.getPlanet(pid)
    species_name = planet.speciesName
    if not species_name:
        species_name = _planned_species().get(pid)
    if species_name:
        return fo.getSpecies(species_name)
    return None


@cache_for_current_turn
def _calculate_get_planet_opinions() -> dict[str, Opinion]:
    universe = fo.getUniverse()
    all_species = [universe.getPlanet(pid).speciesName for pid in get_owned_planets_by_empire()]
    all_species += get_extra_colony_builders()
    all_features = set()
    for species_name in all_species:
        if species_name:
            species = fo.getSpecies(species_name)
            all_features.update(species.likes)
            all_features.update(species.dislikes)

    result = {feature: Opinion(set(), set(), set()) for feature in all_features}
    for pid in get_owned_planets_by_empire():
        species = _planet_species(pid)
        for feature, opinion in result.items():
            if species:
                if feature in species.likes:
                    opinion.likes.add(pid)
                elif feature in species.dislikes:
                    opinion.dislikes.add(pid)
                else:
                    opinion.neutral.add(pid)
            # else: no species -> neutral
            opinion.neutral.add(pid)
    return result


def dislike_factor() -> float:
    """Returns multiplier for dislike effects."""
    # See opinion.macros
    has_liberty = fo.getEmpire().policyAdopted("PLC_LIBERTY")
    has_conformance = fo.getEmpire().policyAdopted("PLC_CONFORMANCE")
    liberty_factor = get_named_real("PLC_LIBERTY_DISLIKE_FACTOR") if has_liberty else 1.0
    conformance_factor = get_named_real("PLC_CONFORMANCE_DISLIKE_FACTOR") if has_conformance else 1.0
    return liberty_factor * conformance_factor


def focus_stability_effect(species: fo.species, focus: str) -> float:
    """How does the focus affect the stability of a planet with the species."""
    result = 0.0
    if focus in species.likes:
        result += STABILITY_PER_LIKED_FOCUS
    if focus in species.dislikes:
        result += STABILITY_PER_LIKED_FOCUS * dislike_factor()
    if focus == FocusType.FOCUS_PROTECTION:
        result += get_game_rule_int("RULE_PROTECTION_FOCUS_STABILITY", 15)
    empire = fo.getEmpire()
    # TODO move policy definitions to AIDependency? Or used an EnumClass like BuildingType?
    if focus == FocusType.FOCUS_INDUSTRY and empire.policyAdopted("PLC_INDUSTRIALISM"):
        result += get_named_real("PLC_INDUSTRIALISM_TARGET_HAPPINESS_FLAT")
    if focus == FocusType.FOCUS_RESEARCH and empire.policyAdopted("PLC_TECHNOCRACY"):
        result += get_named_real("PLC_TECHNOCRACY_TARGET_HAPPINESS_FLAT")
    return result


def stability_with_focus(planet: fo.planet, focus: str) -> float:
    """
    What would the planets target stability be when switched to the given focus.
    Returns -99 if species cannot use the specified focus.
    """
    stability = planet.currentMeterValue(fo.meterType.targetHappiness)
    species = fo.getSpecies(planet.speciesName)
    if focus not in species.foci:
        # The actual value here is not important. If some part of the AI asks for a stability,
        # returning a big negative value here should stop it from considering that focus for anything.
        return -99.0
    return stability - focus_stability_effect(species, planet.focus) + focus_stability_effect(species, focus)


def adjust_liberty(planet: fo.planet, population: float) -> float:
    """
    Adjust liberty research output for changed stability.
    UpdateMeterEstimate calculates liberty based on current stability, not on target stability.
    So this may return a positive or negative adjustment, depending on whether stability goes up or down.
    """
    # This one is relatively easy to calculate. Still it would be better to have an adjustMeterUpdate with
    # modified stability...
    current_stability = planet.currentMeterValue(fo.meterType.happiness)
    target_stability = planet.currentMeterValue(fo.meterType.targetHappiness)
    low = get_named_real("PLC_LIBERTY_MIN_STABILITY")
    high = get_named_real("PLC_LIBERTY_MAX_STABILITY")
    difference = max(low, min(high, target_stability)) - max(low, min(high, current_stability))
    debug(f"adjust_liberty on {planet}: difference={difference}, population={population}")
    return difference * population * get_named_real("PLC_LIBERTY_RESEARCH_BONUS_SCALING")



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_GAIA_TRANS",
    description="GRO_GAIA_TRANS_DESC",
    short_description="GRO_GAIA_TRANS_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=420 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=["GRO_TRANSORG_SENT"],
    unlock=Item(type=UnlockBuilding, name="BLD_GAIA_TRANS"),
    graphic="icons/tech/the_living_planet.png",
)



from logging import error

from AIDependencies import CombatTarget

_issued_errors = set()


def get_allowed_targets(partname: str) -> int:
    """Return the allowed targets for a given hangar or short range weapon."""
    try:
        return CombatTarget.PART_ALLOWED_TARGETS[partname]
    except KeyError:
        if partname not in _issued_errors:
            error(
                "AI has no targeting information for weapon part %s. Will assume any target allowed."
                "Please update CombatTarget.PART_ALLOWED_TARGETS in AIDependencies.py "
            )
            _issued_errors.add(partname)
        return CombatTarget.ANY



import freeOrionAIInterface as fo
import math
from logging import debug
from operator import itemgetter

import AIDependencies
import AIstate
import ColonisationAI
import ExplorationAI
import FleetUtilsAI
import InvasionAI
import MilitaryAI
import PlanetUtilsAI
import ResearchAI
from AIDependencies import INVALID_ID
from aistate_interface import get_aistate
from empire.colony_builders import (
    can_build_colony_for_species,
    can_build_only_sly_colonies,
)
from empire.colony_status import colonies_is_under_attack, colonies_is_under_treat
from EnumsAI import (
    EmpireProductionTypes,
    MissionType,
    PriorityType,
    ShipRoleType,
    get_priority_production_types,
)
from expansion_plans import get_colonisable_outpost_ids, get_colonisable_planet_ids
from freeorion_tools import tech_is_complete
from freeorion_tools.timers import AITimer
from turn_state import (
    get_number_of_colonies,
    get_owned_planets,
    have_asteroids,
    have_gas_giant,
)
from turn_state.design import cur_best_military_design_rating, get_best_ship_info

prioritiees_timer = AITimer("calculate_priorities()")

allottedColonyTargets = 0
allotted_outpost_targets = 0
unmetThreat = 0


def calculate_priorities():
    """Calculates the priorities of the AI player."""
    debug("\n=== Preparing to Calculate Priorities ===")
    prioritiees_timer.start("setting Production Priority")
    aistate = get_aistate()
    # let this one stay fixed & just adjust Research and Influence
    aistate.set_priority(PriorityType.RESOURCE_PRODUCTION, 50)
    # RESOURCE_INFLUENCE priority is calculated in PolicyAI

    debug("\n*** Calculating Research Priority ***\n")
    prioritiees_timer.start("setting Research Priority")
    aistate.set_priority(
        PriorityType.RESOURCE_RESEARCH, _calculate_research_priority()
    )  # TODO: do univ _survey before this

    debug("\n*** Updating Colonization Status ***\n")
    prioritiees_timer.start("Evaluating Colonization Status")
    ColonisationAI.get_colony_fleets()  # TODO sets aistate.colonisablePlanetIDs and many other values used by other modules

    debug("\n*** Updating Invasion Status ***\n")
    prioritiees_timer.start("Evaluating Invasion Status")
    InvasionAI.get_invasion_fleets()  # sets AIstate.opponentPlanetIDs, and AIstate.invasionTargetedPlanetIDs

    debug("\n*** Updating Military Status ***\n")
    prioritiees_timer.start("Evaluating Military Status")
    MilitaryAI.get_military_fleets()

    debug("\n** Calculating Production Priorities ***\n")
    prioritiees_timer.start("reporting Production Priority")
    _calculate_industry_priority()  # purely for reporting purposes
    prioritiees_timer.start("setting Exploration Priority")

    aistate.set_priority(PriorityType.PRODUCTION_EXPLORATION, _calculate_exploration_priority())
    prioritiees_timer.start("setting Colony Priority")
    aistate.set_priority(PriorityType.PRODUCTION_COLONISATION, _calculate_colonisation_priority())
    prioritiees_timer.start("setting Outpost Priority")
    aistate.set_priority(PriorityType.PRODUCTION_OUTPOST, _calculate_outpost_priority())
    prioritiees_timer.start("setting Invasion Priority")
    aistate.set_priority(PriorityType.PRODUCTION_INVASION, _calculate_invasion_priority())
    prioritiees_timer.start("setting Military Priority")
    aistate.set_priority(PriorityType.PRODUCTION_MILITARY, _calculate_military_priority())
    prioritiees_timer.start("setting other priorities")
    aistate.set_priority(PriorityType.PRODUCTION_BUILDINGS, 25)

    prioritiees_timer.stop_print_and_clear()


def _calculate_industry_priority():  # currently only used to print status
    """calculates the demand for industry"""
    universe = fo.getUniverse()
    empire = fo.getEmpire()
    # get current industry production & Target
    industry_production = empire.resourceProduction(fo.resourceType.industry)
    owned_planet_ids = PlanetUtilsAI.get_owned_planets_by_empire()
    planets = (universe.getPlanet(x) for x in owned_planet_ids)
    target_pp = sum(x.currentMeterValue(fo.meterType.targetIndustry) for x in planets)

    # currently, previously set to 50 in calculatePriorities(), this is just for reporting
    industry_priority = get_aistate().get_priority(PriorityType.RESOURCE_PRODUCTION)

    debug("")
    debug(
        f"Industry Production (current/target) : ( {industry_production:.1f} / {target_pp:.1f} ) at turn {fo.currentTurn()}"
    )
    debug("Priority for Industry: %s" % industry_priority)
    return industry_priority


def _calculate_research_priority():  # noqa: C901
    """Calculates the AI empire's demand for research."""
    universe = fo.getUniverse()
    empire = fo.getEmpire()
    current_turn = fo.currentTurn()
    aistate = get_aistate()
    enemies_sighted = aistate.misc.get("enemies_sighted", {})
    recent_enemies = [x for x in enemies_sighted if x > current_turn - 8]

    industry_priority = aistate.get_priority(PriorityType.RESOURCE_PRODUCTION)

    got_algo = tech_is_complete(AIDependencies.LRN_ALGO_ELEGANCE)
    got_quant = tech_is_complete(AIDependencies.LRN_QUANT_NET)
    research_queue_list = ResearchAI.get_research_queue_techs()
    orb_gen_tech = AIDependencies.PRO_ORBITAL_GEN
    got_orb_gen = tech_is_complete(orb_gen_tech)
    mgrav_prod_tech = AIDependencies.PRO_MICROGRAV_MAN
    got_mgrav_prod = tech_is_complete(mgrav_prod_tech)
    # got_solar_gen = tech_is_complete(AIDependencies.PRO_SOL_ORB_GEN)

    milestone_techs = [
        "PRO_ADAPTIVE_AUTOMATION",
        "LRN_DISTRIB_THOUGHT",
        "LRN_QUANT_NET",
        "SHP_WEAPON_2_4",
        "SHP_WEAPON_3_2",
        "SHP_WEAPON_4_2",
    ]
    milestones_done = [mstone for mstone in milestone_techs if tech_is_complete(mstone)]
    debug("Research Milestones accomplished at turn %d: %s" % (current_turn, milestones_done))

    total_pp = empire.productionPoints
    total_rp = empire.resourceProduction(fo.resourceType.research)
    industry_surge = (
        aistate.character.may_surge_industry(total_pp, total_rp)
        and (
            # TODO: having a gas giant does not necessarily mean we want to build a GGG, especially for Sly
            ((orb_gen_tech in research_queue_list[:2] or got_orb_gen) and have_gas_giant())
            or ((mgrav_prod_tech in research_queue_list[:2] or got_mgrav_prod) and have_asteroids())
        )
        and (get_number_of_colonies() < 12)
    )
    # get current industry production & Target
    owned_planet_ids = PlanetUtilsAI.get_owned_planets_by_empire()
    planets = (universe.getPlanet(x) for x in owned_planet_ids)
    target_rp = sum(map(lambda _x: _x.currentMeterValue(fo.meterType.targetResearch), planets))
    galaxy_is_sparse = ColonisationAI.galaxy_is_sparse()
    enemies_sighted = aistate.misc.get("enemies_sighted", {})

    style_index = aistate.character.preferred_research_cutoff([0, 1])
    if aistate.character.may_maximize_research():
        style_index += 1

    cutoff_sets = [[25, 45, 70, 110], [30, 45, 70, 150], [25, 40, 80, 160]]
    cutoffs = cutoff_sets[style_index]
    settings = [[1.3, 0.7, 0.5, 0.4, 0.35], [1.4, 0.8, 0.6, 0.5, 0.35], [1.4, 0.8, 0.6, 0.5, 0.4]][style_index]

    if (
        (current_turn < cutoffs[0])
        or (not got_algo)
        or ((style_index == 0) and not got_orb_gen and (current_turn < cutoffs[1]))
    ):
        research_priority = (
            settings[0] * industry_priority
        )  # high research at beginning of game to get easy gro tech and to get research booster Algotrithmic Elegance
    elif (not got_orb_gen) or (current_turn < cutoffs[1]):
        research_priority = settings[1] * industry_priority  # med-high research
    elif current_turn < cutoffs[2]:
        research_priority = settings[2] * industry_priority  # med-high industry
    elif current_turn < cutoffs[3]:
        research_priority = settings[3] * industry_priority  # med-high industry
    else:
        research_queue = list(empire.researchQueue)
        research_priority = settings[4] * industry_priority  # high industry , low research
        if len(research_queue) == 0:
            research_priority = 0  # done with research
        elif len(research_queue) < 5 and research_queue[-1].allocation > 0:
            research_priority = len(research_queue) * 0.01 * industry_priority  # barely not done with research
        elif len(research_queue) < 10 and research_queue[-1].allocation > 0:
            research_priority = (4 + 2 * len(research_queue)) * 0.01 * industry_priority  # almost done with research
        elif len(research_queue) < 20 and research_queue[int(len(research_queue) // 2)].allocation > 0:
            research_priority *= 0.7  # closing in on end of research
    if industry_surge:
        if galaxy_is_sparse and not any(enemies_sighted):
            research_priority *= 0.5
        else:
            research_priority *= 0.8

    if (tech_is_complete("SHP_WEAPON_2_4") or tech_is_complete("SHP_WEAPON_4_1")) and tech_is_complete(
        AIDependencies.PRO_AUTO_1
    ):
        # industry_factor = [ [0.25, 0.2], [0.3, 0.25], [0.3, 0.25] ][style_index ]
        # researchPriority = min(researchPriority, industry_factor[got_solar_gen]*industryPriority)
        research_priority *= 0.9
    if got_quant:
        research_priority = min(research_priority + 0.1 * industry_priority, research_priority * 1.3)
    research_priority = int(research_priority)
    debug(f"Research Production (current/target) : ( {total_rp:.1f} / {target_rp:.1f} )")
    debug(
        "Priority for Research: %d (new target ~ %d RP)"
        % (research_priority, total_pp * research_priority / industry_priority)
    )

    if len(enemies_sighted) < (2 + current_turn / 20.0):  # TODO: adjust for colonisation priority
        research_priority *= 1.2
    if (current_turn > 20) and (len(recent_enemies) > 3):
        research_priority *= 0.8
    return research_priority


def _calculate_exploration_priority():
    """Calculates the demand for scouts by unexplored systems."""
    empire = fo.getEmpire()
    num_unexplored_systems = len(ExplorationAI.border_unexplored_system_ids)
    aistate = get_aistate()
    num_scouts = sum(
        [
            aistate.fleetStatus.get(fid, {}).get("nships", 0)
            for fid in FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.EXPLORATION)
        ]
    )
    production_queue = empire.productionQueue
    queued_scout_ships = 0
    for queue_index in range(0, len(production_queue)):
        element = production_queue[queue_index]
        if element.buildType == EmpireProductionTypes.BT_SHIP:
            if aistate.get_ship_role(element.designID) == ShipRoleType.CIVILIAN_EXPLORATION:
                queued_scout_ships += element.remaining * element.blocksize

    mil_ships = MilitaryAI.get_num_military_ships()
    # intent of the following calc is essentially
    # new_scouts_needed = min(need_cap_A, need_cap_B, base_need) - already_got_or_queued
    # where need_cap_A is to help prevent scouting needs from swamping military needs, and
    # need_cap_B is to help regulate investment into scouting while the empire is small.
    # These caps could perhaps instead be tied more directly to military priority and
    # total empire production.
    desired_number_of_scouts = int(
        min(4 + mil_ships / 5, 4 + fo.currentTurn() / 50.0, 2 + num_unexplored_systems**0.5)
    )
    scouts_needed = max(0, desired_number_of_scouts - (num_scouts + queued_scout_ships))
    exploration_priority = int(40 * scouts_needed)

    debug("")
    debug("Number of Scouts: %s" % num_scouts)
    debug("Number of Unexplored systems: %s" % num_unexplored_systems)
    debug("Military size: %s" % mil_ships)
    debug("Priority for scouts: %s" % exploration_priority)

    return exploration_priority


def _calculate_colonisation_priority():
    """Calculates the demand for colony ships by colonisable planets."""
    global allottedColonyTargets
    aistate = get_aistate()
    enemies_sighted = aistate.misc.get("enemies_sighted", {})
    galaxy_is_sparse = ColonisationAI.galaxy_is_sparse()
    total_pp = fo.getEmpire().productionPoints
    num_colonies = get_number_of_colonies()
    colony_growth_barrier = aistate.character.max_number_colonies()
    if num_colonies > colony_growth_barrier:
        return 0.0
    colony_cost, turns_to_build = ColonisationAI.colony_pod_cost_turns()

    mil_prio = aistate.get_priority(PriorityType.PRODUCTION_MILITARY)
    allotted_portion = [[[0.6, 0.8], [0.3, 0.4]], [[0.8, 0.9], [0.4, 0.6]]][galaxy_is_sparse][any(enemies_sighted)]
    allotted_portion = aistate.character.preferred_colonization_portion(allotted_portion)
    # if ( get_aistate().get_priority(AIPriorityType.PRIORITY_PRODUCTION_COLONISATION)
    # > 2 * get_aistate().get_priority(AIPriorityType.PRIORITY_PRODUCTION_MILITARY)):
    # allotted_portion *= 1.5
    if mil_prio < 100:
        allotted_portion *= 2
    elif mil_prio < 200:
        allotted_portion *= 1.5
    elif fo.currentTurn() > 100:
        allotted_portion *= 0.75 ** (num_colonies / 10.0)
    # allottedColonyTargets = 1+ int(fo.currentTurn()/50)
    allottedColonyTargets = 1 + int(total_pp * turns_to_build * allotted_portion / colony_cost)
    outpost_prio = aistate.get_priority(PriorityType.PRODUCTION_OUTPOST)

    # if have no SP_SLY, and have any outposts to build, don't build colony ships TODO: make more complex assessment
    if not can_build_colony_for_species("SP_SLY") and outpost_prio > 0:
        return 0.0
    min_score = ColonisationAI.MINIMUM_COLONY_SCORE
    minimal_top = min_score + 2  # one more than the conditional floor set by ColonisationAI.revise_threat_factor()
    minimal_opportunities = [
        species_name
        for (_, (score, species_name)) in get_colonisable_planet_ids().items()
        if min_score < score <= minimal_top
    ]
    decent_opportunities = [
        species_name for (_, (score, species_name)) in get_colonisable_planet_ids().items() if score > minimal_top
    ]
    minimal_planet_factor = 0.2  # count them for something, but not much
    num_colonisable_planet_ids = len(decent_opportunities) + minimal_planet_factor * len(minimal_opportunities)
    if num_colonisable_planet_ids == 0:
        return 1

    colony_ship_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.COLONISATION)
    num_colony_ships = len(FleetUtilsAI.extract_fleet_ids_without_mission_types(colony_ship_ids))
    colonisation_priority = (
        60 * (1.0 + num_colonisable_planet_ids - num_colony_ships) / (num_colonisable_planet_ids + 1)
    )

    if can_build_only_sly_colonies():
        colonisation_priority *= 2
    elif can_build_colony_for_species("SP_SLY"):
        colony_opportunities = minimal_opportunities + decent_opportunities
        colonisation_priority *= (1.0 + colony_opportunities.count("SP_SLY")) / len(colony_opportunities)

    if colonisation_priority < 1:
        return 0
    return colonisation_priority


def _calculate_outpost_priority():
    """Calculates the demand for outpost ships by colonisable planets."""
    global allotted_outpost_targets
    base_outpost_cost = AIDependencies.OUTPOST_POD_COST

    aistate = get_aistate()
    enemies_sighted = aistate.misc.get("enemies_sighted", {})
    galaxy_is_sparse = ColonisationAI.galaxy_is_sparse()
    total_pp = fo.getEmpire().productionPoints
    colony_growth_barrier = aistate.character.max_number_colonies()
    if get_number_of_colonies() > colony_growth_barrier:
        return 0.0
    mil_prio = aistate.get_priority(PriorityType.PRODUCTION_MILITARY)

    not_sparse, enemy_unseen = 0, 0
    is_sparse, enemy_seen = 1, 1
    allotted_portion = {
        (not_sparse, enemy_unseen): (0.6, 0.8),
        (not_sparse, enemy_seen): (0.3, 0.4),
        (is_sparse, enemy_unseen): (0.8, 0.9),
        (is_sparse, enemy_seen): (0.3, 0.4),
    }[(galaxy_is_sparse, any(enemies_sighted))]
    allotted_portion = aistate.character.preferred_outpost_portion(allotted_portion)
    if mil_prio < 100:
        allotted_portion *= 2
    elif mil_prio < 200:
        allotted_portion *= 1.5
    allotted_outpost_targets = 1 + int(total_pp * 3 * allotted_portion / base_outpost_cost)

    num_outpost_targets = len(
        [
            pid
            for (pid, (score, specName)) in get_colonisable_outpost_ids().items()
            if score > max(1.0 * base_outpost_cost / 3.0, ColonisationAI.MINIMUM_COLONY_SCORE)
        ][:allotted_outpost_targets]
    )
    if num_outpost_targets == 0 or not tech_is_complete(AIDependencies.OUTPOSTING_TECH):
        return 0

    outpost_ship_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.OUTPOST)
    num_outpost_ships = len(FleetUtilsAI.extract_fleet_ids_without_mission_types(outpost_ship_ids))
    outpost_priority = (50.0 * (num_outpost_targets - num_outpost_ships)) / num_outpost_targets

    # If the only possible race for colonization is SP_SLY (early stage of game before conquers and exobots)
    # discourage early outposting due to supply and stealth considerations they are best off
    # using colony ships until they have other colonizers (and more established military)
    if can_build_only_sly_colonies():
        outpost_priority /= 3.0

    # print
    # print "Number of Outpost Ships : " + str(num_outpost_ships)
    # print "Number of Colonisable outposts: " + str(num_outpost_planet_ids)
    debug("Priority for outpost ships: %s" % outpost_priority)

    if outpost_priority < 1:
        return 0
    return outpost_priority


def _calculate_invasion_priority():  # noqa: C901
    """Calculates the demand for troop ships by opponent planets."""

    aistate = get_aistate()
    if not aistate.character.may_invade():
        return 0

    empire = fo.getEmpire()
    enemies_sighted = aistate.misc.get("enemies_sighted", {})
    multiplier = 1
    colony_growth_barrier = aistate.character.max_number_colonies()
    if get_number_of_colonies() > colony_growth_barrier:
        return 0.0

    if len(aistate.colonisablePlanetIDs) > 0:
        best_colony_score = max(2, next(iter(aistate.colonisablePlanetIDs.items()))[1][0])
    else:
        best_colony_score = 2

    total_val = 0
    troops_needed = 0
    for pid, pscore, trp in AIstate.invasionTargets[: allotted_invasion_targets()]:
        if pscore > best_colony_score:
            multiplier += 1
            total_val += 2 * pscore
        else:
            total_val += pscore
        troops_needed += trp + 4  # ToDo: This seems like it could be improved by some dynamic calculation of buffer

    if total_val == 0:
        return 0

    production_queue = empire.productionQueue
    queued_troop_capacity = 0
    for queue_index in range(0, len(production_queue)):
        element = production_queue[queue_index]
        if element.buildType == EmpireProductionTypes.BT_SHIP:
            if aistate.get_ship_role(element.designID) in [ShipRoleType.MILITARY_INVASION, ShipRoleType.BASE_INVASION]:
                design = fo.getShipDesign(element.designID)
                queued_troop_capacity += element.remaining * element.blocksize * design.troopCapacity
    _, best_design, _ = get_best_ship_info(PriorityType.PRODUCTION_INVASION)
    if best_design:
        troops_per_best_ship = best_design.troopCapacity
    else:
        return 1e-6  # if we can not build troop ships, we don't want to build (non-existing) invasion ships

    # don't count troop bases here as these cannot be redeployed after misplaning
    # troopFleetIDs = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.INVASION)\
    #                 + FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.ORBITAL_INVASION)
    troop_fleet_ids = FleetUtilsAI.get_empire_fleet_ids_by_role(MissionType.INVASION)
    total_troop_capacity = sum([FleetUtilsAI.count_troops_in_fleet(fid) for fid in troop_fleet_ids])
    troop_ships_needed = math.ceil(
        (troops_needed - (total_troop_capacity + queued_troop_capacity)) / troops_per_best_ship
    )

    # invasion_priority = max( 10+ 200*max(0, troop_ships_needed ) , int(0.1* total_val) )
    invasion_priority = multiplier * (30 + 150 * max(0, troop_ships_needed))
    if not colonies_is_under_attack():
        if not colonies_is_under_treat():
            invasion_priority *= 2.0
        else:
            invasion_priority *= 1.5
    if not enemies_sighted:
        invasion_priority *= 1.5

    if invasion_priority < 0:
        return 0

    return invasion_priority * aistate.character.invasion_priority_scaling()


def allotted_invasion_targets():
    return min(1 + int(fo.currentTurn() // 50), 3)


def _calculate_military_priority():  # noqa: C901
    """Calculates the demand for military ships by military targeted systems."""
    global unmetThreat

    universe = fo.getUniverse()
    capital_id = PlanetUtilsAI.get_capital()
    if capital_id is None or capital_id == INVALID_ID:
        return 0  # no capitol (not even a capitol-in-the-making), means can't produce any ships

    have_l1_weaps = (
        tech_is_complete("SHP_WEAPON_1_4")
        or (tech_is_complete("SHP_WEAPON_1_3") and tech_is_complete("SHP_MIL_ROBO_CONT"))
        or tech_is_complete("SHP_WEAPON_2_1")
        or tech_is_complete("SHP_WEAPON_4_1")
    )
    have_l2_weaps = tech_is_complete("SHP_WEAPON_2_3") or tech_is_complete("SHP_WEAPON_4_1")
    aistate = get_aistate()
    enemies_sighted = aistate.misc.get("enemies_sighted", {})

    target_planet_ids = (
        [pid for pid, pscore, trp in AIstate.invasionTargets[: allotted_invasion_targets()]]
        + [pid for pid, pscore in list(get_colonisable_planet_ids(True).items())[:allottedColonyTargets]]
        + [pid for pid, pscore in list(get_colonisable_outpost_ids(True).items())[:allottedColonyTargets]]
    )

    my_systems = set(get_owned_planets())
    target_systems = set(PlanetUtilsAI.get_systems(target_planet_ids))

    cur_ship_rating = cur_best_military_design_rating()
    current_turn = fo.currentTurn()
    ships_needed = 0
    defense_ships_needed = 0
    ships_needed_allocation = []
    for sys_id in my_systems.union(target_systems):
        status = aistate.systemStatus.get(sys_id, {})
        my_rating = status.get("myFleetRating", 0)
        my_defenses = status.get("mydefenses", {}).get("overall", 0)
        base_monster_threat = status.get("monsterThreat", 0)
        # scale monster threat so that in early - mid game big monsters don't over-drive military production
        monster_threat = base_monster_threat
        if current_turn > 200:
            pass
        elif current_turn > 100:
            if base_monster_threat >= 2000:
                monster_threat = 2000 + (current_turn / 100.0 - 1) * (base_monster_threat - 2000)
        elif current_turn > 30:
            if base_monster_threat >= 2000:
                monster_threat = 0
        else:
            if base_monster_threat > 200:
                monster_threat = 0
        if sys_id in my_systems:
            threat_root = (
                status.get("fleetThreat", 0) ** 0.5
                + 0.8 * status.get("max_neighbor_threat", 0) ** 0.5
                + 0.2 * status.get("neighborThreat", 0) ** 0.5
                + monster_threat**0.5
                + status.get("planetThreat", 0) ** 0.5
            )
        else:
            threat_root = (
                status.get("fleetThreat", 0) ** 0.5 + monster_threat**0.5 + status.get("planetThreat", 0) ** 0.5
            )
        ships_needed_here = math.ceil(
            (max(0, (threat_root - (my_rating**0.5 + my_defenses**0.5))) ** 2) / cur_ship_rating
        )
        ships_needed += ships_needed_here
        ships_needed_allocation.append((universe.getSystem(sys_id), ships_needed_here))
        if sys_id in my_systems:
            defense_ships_needed += ships_needed_here

    part1 = min(1 * fo.currentTurn(), 40)
    part2 = max(0, int(75 * ships_needed))
    military_priority = part1 + part2
    if not have_l1_weaps:
        military_priority /= 2.0
    elif not (have_l2_weaps and enemies_sighted):
        military_priority /= 1.5
    fmt_string = "Calculating Military Priority:  min(t,40) + max(0,75 * ships_needed) \n\t  Priority: %d  \t ships_needed: %d \t defense_ships_needed: %d \t curShipRating: %.0f \t l1_weaps: %s \t enemies_sighted: %s"
    debug(
        fmt_string
        % (military_priority, ships_needed, defense_ships_needed, cur_ship_rating, have_l1_weaps, enemies_sighted)
    )
    debug("Source of milship demand: %s" % ships_needed_allocation)

    military_priority *= aistate.character.military_priority_scaling()
    return max(military_priority, 0)


def _calculate_top_production_queue_priority():
    """Calculates the top production queue priority."""
    production_queue_priorities = {}
    aistate = get_aistate()
    for priorityType in get_priority_production_types():
        production_queue_priorities[priorityType] = aistate.get_priority(priorityType)

    sorted_priorities = sorted(production_queue_priorities.items(), key=itemgetter(1), reverse=True)
    top_production_queue_priority = -1
    for evaluationPair in sorted_priorities:
        if top_production_queue_priority < 0:
            top_production_queue_priority = evaluationPair[0]

    return top_production_queue_priority



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="GRO_NANO_CYBERNET",
    description="GRO_NANO_CYBERNET_DESC",
    short_description="SHIP_PART_UNLOCK_SHORT_DESC",
    category="GROWTH_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_GROWTH_CATEGORY"],
    prerequisites=["GRO_NANOTECH_MED"],
    unlock=[Item(type=UnlockShipPart, name="GT_TROOP_POD_2"), Item(type=UnlockPolicy, name="PLC_AUGMENTATION")],
    graphic="icons/tech/nanotech_cybernetics.png",
)



from focs._species import *
from species.species_macros.empire_opinions import COMMON_OPINION_EFFECTS
from species.species_macros.env import DESERT_NARROW_EP, NARROW_EP
from species.species_macros.focus import (
    HAS_ADVANCED_FOCI,
    HAS_GROWTH_FOCUS,
    HAS_INDUSTRY_FOCUS,
)
from species.species_macros.happiness import AVERAGE_HAPPINESS
from species.species_macros.industry import BAD_INDUSTRY
from species.species_macros.influence import NO_INFLUENCE
from species.species_macros.native_fortification import DEFAULT_NATIVE_DEFENSE
from species.species_macros.population import BAD_POPULATION
from species.species_macros.research import NO_RESEARCH
from species.species_macros.shields import STANDARD_SHIP_SHIELDS
from species.species_macros.stockpile import GOOD_STOCKPILE
from species.species_macros.supply import AVERAGE_SUPPLY
from species.species_macros.troops import GREAT_DEFENSE_TROOPS, GREAT_OFFENSE_TROOPS

Species(
    name="SP_KHAKTURIAN",
    description="SP_KHAKTURIAN_DESC",
    gameplay_description="SP_KHAKTURIAN_GAMEPLAY_DESC",
    native=True,
    can_produce_ships=True,
    tags=[
        "PHOTOTROPHIC",
        "BAD_POPULATION",
        "BAD_INDUSTRY",
        "NO_RESEARCH",
        "NO_INFLUENCE",
        "GOOD_STOCKPILE",
        "GREAT_OFFENSE_TROOPS",
        "PEDIA_PHOTOTROPHIC_SPECIES_CLASS",
    ],
    foci=[
        HAS_INDUSTRY_FOCUS,
        # [[HAS_RESEARCH_FOCUS]]*/    // dumb plants after all
        # [[HAS_INFLUENCE_FOCUS]]*/   // see above
        HAS_GROWTH_FOCUS,
        *HAS_ADVANCED_FOCI,
    ],
    defaultfocus="FOCUS_PROTECTION",
    likes=[
        "BLD_LIGHTHOUSE",  # if there were a greenhouse they would like that too
        "BLD_STOCKPILING_CENTER",  # why, to store all the seeds for the next rainy season of course
        "FOCUS_GROWTH",  # what else would a plant want?
        "FOCUS_PROTECTION",  # yep, that one too.. happily bristling with thorns, soaking sun, la la la
        "HEAD_ON_A_SPIKE_SPECIAL",  # we likes the spikes
        "PLC_POPULATION",  # grow, grow, grow your load, gently down the ..
        "TIDAL_LOCK_SPECIAL",  # eternal daylight yeah! nightsiders? not a problem for us prickly plants we got no blood
    ],
    dislikes=[  # dislikes are each on their own single line of code to facilitate commenting and merging,
        "CLOUD_COVER_MASTER_SPECIAL",  # stop blockin the sun over there
        "CLOUD_COVER_SLAVE_SPECIAL",  # same. emphasis.
        "PLC_NO_GROWTH",  # dumb plants, but looove to grow
        "VOLCANIC_ASH_MASTER_SPECIAL",  # no sunblock pls
        "VOLCANIC_ASH_SLAVE_SPECIAL",  # ...
        # see also: WORLDTREE.focs.txt, where only the SP_KHAKTURIAN are explicitely excluded from worldree bonus effects,
    ],
    effectsgroups=[
        *BAD_INDUSTRY,
        NO_RESEARCH,
        *NO_INFLUENCE,
        *GOOD_STOCKPILE,
        *BAD_POPULATION,
        *AVERAGE_HAPPINESS,
        COMMON_OPINION_EFFECTS("SP_KHAKTURIAN"),
        *AVERAGE_SUPPLY,
        *GREAT_OFFENSE_TROOPS,
        *GREAT_DEFENSE_TROOPS,
        # not for description,
        *DEFAULT_NATIVE_DEFENSE,
        *STANDARD_SHIP_SHIELDS,
        NARROW_EP,
    ],
    environments=DESERT_NARROW_EP,
    graphic="icons/species/khakturian.png",
)



import os
from logging import debug, error

from stub_generator.generate_stub import make_stub
from stub_generator.interface_inspector import get_module_info
from stub_generator.stub_generator.rtype.mapper import report_unused_mapping


def generate_stub(obj, instances, classes_to_ignore, path):
    """
    Inspect interface and generate stub. Writes its logs to freeoriond.log.

    :param obj: main interface module (freeOrionAIInterface for AI)
    :param instances:  list of instances, required to get more detailed information about them
    :param classes_to_ignore: classes that should not to be reported when check for missed instances done.
                              this argument required because some classes present in interface
                              but have no methods, to get their instances.
    :param path: relative path from python folder
    """
    debug("=" * 20)
    debug("\n\nStart generating skeleton for %s\n\n" % obj.__name__)
    debug("=" * 20)
    python_folder_path = os.path.normpath(os.path.join(os.path.dirname(__file__), ".."))
    result_folder = os.path.join(python_folder_path, path)
    result_path = os.path.join(result_folder, "%s.pyi" % obj.__name__)
    classes, enums, functions, instances = get_module_info(obj, instances)
    make_stub(classes, enums, functions, instances, result_path, classes_to_ignore)

    report_unused_mapping()

    debug("=" * 20)
    error("Skeleton was created, please exit to the menu")
    debug("Skeleton written to %s" % result_path)
    debug("=" * 20)



import os
import sys
from collections import OrderedDict as odict
from configparser import ConfigParser
from pathlib import Path
from typing import Any, Optional

DEFAULT_SUB_DIR = Path("AI", "default")
CONFIG_DEFAULT_FILE = "config.ini"

# CONFIG KEYS
TIMER_SECTION = "Timers"
TIMERS_USE_TIMERS = "timers_to_log"
HANDLERS = "handlers"


flat_options = odict()


def _get_option_file(config_dir: Path) -> Optional[Path]:
    # hack to allow lunch this code separately to dump default config
    ai_path = _get_default_file_path(config_dir)
    option_file = CONFIG_DEFAULT_FILE
    if ai_path and option_file:
        return ai_path / option_file
    else:
        return None


def _get_default_file_path(config_dir: Path) -> Path:
    config_default_dir = config_dir / DEFAULT_SUB_DIR
    try:
        if config_dir.is_dir() and not config_default_dir.is_dir():
            config_default_dir.mkdir(parents=True)
    except OSError:
        sys.stderr.write(f"AI Config Error: could not create path {config_default_dir}\n")
        return config_dir
    return config_default_dir


def _get_preset_default_ai_options():
    """
    returns preset default options for AI; an ordered dict of dicts.
    each sub-dict corresponds to a config file section
    :return:
    """
    return odict(
        [
            (
                TIMER_SECTION,
                odict([(TIMERS_USE_TIMERS, False)]),
            ),
            ("main", odict([(HANDLERS, "")])),  # module names in handler directory, joined by space
        ]
    )


def _create_default_config_file(path: Path) -> ConfigParser:
    """
    Writes default config to file.
    """
    config = ConfigParser()
    presets = _get_preset_default_ai_options()
    for section, entries in presets.items():
        config.add_section(section)
        for k, v in entries.items():
            config.set(section, k, str(v))
    if path:
        try:
            with open(path, "w") as configfile:
                config.write(configfile)
            print(f"default config is dumped to {path}")
        except OSError:
            sys.stderr.write(f"AI Config Error: could not write default config {path}\n")
    return config


def check_bool(option):
    return str(option).lower() in ["1", "on", "true", "yes"]


def get_option_dict():
    """
    Return options for AI
    :return: ordered dict of options
    """
    return flat_options


def parse_config(option_string: Any, config_dir: str):
    config_dir = Path(config_dir)

    if option_string is not None and not isinstance(option_string, str):
        # probably called by unit test
        print("Specified option string is not a string: ", option_string, file=sys.stderr)
        return

    # get defaults; check if don't already exist and can write
    default_file = _get_option_file(config_dir)
    if os.path.exists(default_file):
        config = ConfigParser()
        config.read([default_file])
        print("AI Config read from default file(s): %s" % default_file)
    else:
        try:
            config = _create_default_config_file(default_file)
        except OSError:
            sys.stderr.write("AI Config: default file is not present and not writable at location %s\n" % default_file)
            config = _create_default_config_file(config_dir / CONFIG_DEFAULT_FILE)

    if option_string:
        config_files = [option_string]
        configs_read = config.read(config_files)
        print("AI Config read config file(s): %s" % configs_read)
        if len(configs_read) != len(config_files):
            sys.stderr.write(
                "AI Config Error; could NOT read config file(s): %s\n"
                % list(set(config_files).difference(configs_read))
            )
    for section in config.sections():
        for k, v in config.items(section):
            flat_options[k] = v



from focs._effects import (
    EffectsGroup,
    Focus,
    Happiness,
    NamedReal,
    OwnedBy,
    Planet,
    SetTargetResearch,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import RESEARCH_PER_POP, TECH_COST_MULTIPLIER
from macros.priorities import TARGET_AFTER_SCALING_PRIORITY

Tech(
    name="LRN_QUANT_NET",
    description="LRN_QUANT_NET_DESC",
    short_description="RESEARCH_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=300 * TECH_COST_MULTIPLIER,
    researchturns=6,
    tags=["PEDIA_LEARNING_CATEGORY"],
    prerequisites=["LRN_NDIM_SUBSPACE"],
    effectsgroups=[
        EffectsGroup(
            scope=Planet()
            & OwnedBy(empire=Source.Owner)
            & Focus(type=["FOCUS_RESEARCH"])
            & Happiness(low=NamedReal(name="LRN_QUANT_NET_MIN_STABILITY", value=10)),
            priority=TARGET_AFTER_SCALING_PRIORITY,
            effects=SetTargetResearch(
                value=Value
                + Target.Population
                * NamedReal(name="LRN_QUANT_NET_TARGET_RESEARCH_PERPOP", value=(0.5 * RESEARCH_PER_POP))
            ),
        )
    ],
    graphic="icons/tech/quantum_networking.png",
)



from focs._effects import (
    Abs,
    Adequate,
    Capital,
    CurrentTurn,
    DesignHasPart,
    EffectsGroup,
    EmpireStockpile,
    Focus,
    GalaxyMaxAIAggression,
    Happiness,
    HasEmpireStockpile,
    HasSpecies,
    Homeworld,
    Hostile,
    Influence,
    IsHuman,
    IsSource,
    LocalCandidate,
    MaxOf,
    MinOf,
    NamedReal,
    NamedRealLookup,
    OwnedBy,
    PartCapacity,
    Planet,
    Poor,
    SetCapacity,
    SetMaxTroops,
    SetRebelTroops,
    SetTargetHappiness,
    SetTargetInfluence,
    SetTroops,
    Ship,
    Source,
    StatisticCount,
    Target,
    Unowned,
    Value,
)
from macros.base_prod import TROOPS_PER_POP
from macros.priorities import (
    AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
    TARGET_AFTER_2ND_SCALING_PRIORITY,
    TARGET_EARLY_BEFORE_SCALING_PRIORITY,
    TARGET_SCALING_PRIORITY,
)

_UNSTABLE_REBEL_TROOPS = [
    EffectsGroup(
        scope=IsSource,
        activation=~Unowned & ~Happiness(low=0),
        accountinglabel="INSTABILITY_REBELLION",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        # more negative stability and more population generates more rebels each turn
        # TODO: scale with some species trait?
        effects=SetRebelTroops(value=Value + Abs(float, Target.Happiness) * (Target.Population / 5.0) ** 0.5),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Unowned & ~HasEmpireStockpile(empire=Source.Owner, resource=Influence, low=0) & ~Capital
        # TODO: some species trait?
        ,
        accountinglabel="PARTISANS_REBELLION",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetHappiness(
            value=Value - 0.5 * Abs(float, EmpireStockpile(empire=Source.Owner, resource=Influence)) ** 0.5
        ),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Unowned & ~HasEmpireStockpile(empire=Source.Owner, resource=Influence, low=0) & ~Capital,
        accountinglabel="INFLUENCE_DEBT_DECAY",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetTargetInfluence(
            value=Value
            + 0.5
            * Abs(float, EmpireStockpile(empire=Source.Owner, resource=Influence)) ** 0.5
            / MaxOf(float, 1.0, StatisticCount(float, condition=Planet() & HasSpecies() & OwnedBy(empire=Source.Owner)))
        ),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Adequate]),
        accountinglabel="DEPENDENCE_ON_IMPERIAL_SUPPORT_DUE_TO_ENVIRONMENT",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetRebelTroops(value=MaxOf(float, 0.0, Abs(float, Value / 1.5))),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Poor]),
        accountinglabel="DEPENDENCE_ON_IMPERIAL_SUPPORT_DUE_TO_ENVIRONMENT",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetRebelTroops(value=MaxOf(float, 0.0, Abs(float, Value / 2.0))),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet(environment=[Hostile]),
        accountinglabel="DEPENDENCE_ON_IMPERIAL_SUPPORT_DUE_TO_ENVIRONMENT",
        priority=TARGET_SCALING_PRIORITY,
        effects=SetRebelTroops(value=MaxOf(float, 0.0, Abs(float, Value / 4.0))),
    ),
]

_PROTECTION_FOCUS_TROOPS = [
    EffectsGroup(  # double max troops
        scope=IsSource,
        activation=Planet() & Focus(type=["FOCUS_PROTECTION"]),
        stackinggroup="FOCUS_PROTECTION_TROOPS_STACK",
        priority=TARGET_AFTER_2ND_SCALING_PRIORITY,
        effects=SetMaxTroops(value=MaxOf(float, Value, Value * 2)),
        # accountinglabel="FOCUS_PROTECTION_LABEL",
    ),
    EffectsGroup(  # increase troop growth rate
        scope=IsSource,
        activation=Planet()
        & Focus(type=["FOCUS_PROTECTION"])
        & (LocalCandidate.LastTurnConquered < CurrentTurn)
        & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
        accountinglabel="FOCUS_PROTECTION_LABEL",
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetTroops(
            value=MinOf(
                float, Value(Target.MaxTroops), Value + MaxOf(float, Target.MaxTroops, Target.Population) ** 0.5
            )
        ),
    ),
]


_BASIC_DEFENSE_TROOPS = [
    EffectsGroup(
        scope=IsSource,
        activation=Homeworld() & ~Unowned,
        stackinggroup="HOMEWORLD_TROOPS_STACK",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetMaxTroops(value=Value + 4),
        # accountinglabel="HOMEWORLD_LABEL",
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & ~Unowned,
        accountinglabel="IMPERIAL_GARRISON_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetMaxTroops(value=Value + NamedRealLookup(name="IMPERIAL_GARRISON_MAX_TROOPS_FLAT")),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet() & Unowned,
        stackinggroup="BASIC_TROOPS_STACK",
        accountinglabel="INDEPENDENT_TROOP_LABEL",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetMaxTroops(value=Value + 10),
    ),
    EffectsGroup(
        scope=IsSource,
        activation=Planet(),
        stackinggroup="POPULATION_TROOPS_STACK",
        accountinglabel="DEF_ROOT_DEFENSE",
        priority=TARGET_EARLY_BEFORE_SCALING_PRIORITY,
        effects=SetMaxTroops(
            value=Value
            + Target.Population * NamedReal(name="BASIC_DEFENSE_TROOPS_MAX_TROOPS_PERPOP", value=1.0 * TROOPS_PER_POP)
        ),
    ),
    EffectsGroup(  # gives human bonuses when AI Aggression set to Beginner
        scope=IsSource,
        activation=IsHuman & (GalaxyMaxAIAggression == 0) & Planet(),  # human player, not human species
        accountinglabel="DIFFICULTY",
        effects=SetMaxTroops(value=MaxOf(float, 6, Value * 2)),
    ),
    EffectsGroup(  # base troops regeneration
        scope=IsSource,
        activation=Planet()
        & (LocalCandidate.LastTurnConquered < CurrentTurn)
        & (LocalCandidate.LastTurnAttackedByShip < CurrentTurn),
        priority=AFTER_ALL_TARGET_MAX_METERS_PRIORITY,
        effects=SetTroops(
            value=MinOf(
                float, Value(Target.MaxTroops), Value + NamedReal(name="BASIC_DEFENSE_TROOPS_TROOPREGEN_FLAT", value=1)
            )
        ),
    ),
]

NO_DEFENSE_TROOPS = [
    EffectsGroup(description="NO_DEFENSE_TROOPS_DESC", scope=IsSource, activation=None, effects=SetMaxTroops(value=0)),
    *_UNSTABLE_REBEL_TROOPS,
]


BAD_DEFENSE_TROOPS = [
    *_BASIC_DEFENSE_TROOPS,
    EffectsGroup(
        description="BAD_DEFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Planet(),
        effects=SetMaxTroops(value=Value * 0.5),
    ),
    *_UNSTABLE_REBEL_TROOPS,
    *_PROTECTION_FOCUS_TROOPS,
]

AVERAGE_DEFENSE_TROOPS = [
    *_BASIC_DEFENSE_TROOPS,
    *_UNSTABLE_REBEL_TROOPS,
    *_PROTECTION_FOCUS_TROOPS,
]


GOOD_DEFENSE_TROOPS = [
    *_BASIC_DEFENSE_TROOPS,
    EffectsGroup(
        description="GOOD_DEFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Planet(),
        effects=SetMaxTroops(value=Value * 1.5),
    ),
    *_UNSTABLE_REBEL_TROOPS,
    *_PROTECTION_FOCUS_TROOPS,
]


GREAT_DEFENSE_TROOPS = [
    *_BASIC_DEFENSE_TROOPS,
    EffectsGroup(
        description="GREAT_DEFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Planet(),
        effects=SetMaxTroops(value=Value * 2),
    ),
    *_UNSTABLE_REBEL_TROOPS,
    *_PROTECTION_FOCUS_TROOPS,
]

ULTIMATE_DEFENSE_TROOPS = [
    *_BASIC_DEFENSE_TROOPS,
    EffectsGroup(
        description="ULTIMATE_DEFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Planet(),
        effects=SetMaxTroops(value=Value * 3),
    ),
    *_UNSTABLE_REBEL_TROOPS,
    *_PROTECTION_FOCUS_TROOPS,
]


NO_OFFENSE_TROOPS = [
    EffectsGroup(
        description="NO_OFFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Ship & (DesignHasPart(name="GT_TROOP_POD") | DesignHasPart(name="GT_TROOP_POD_2")),
        stackinggroup="NO_OFFENSIVE_TROOPS_STACK",
        accountinglabel="NO_OFFENSIVE_TROOPS_LABEL",
        effects=[
            SetCapacity(partname="GT_TROOP_POD", value=0.0),
            SetCapacity(partname="GT_TROOP_POD_2", value=0.0),
        ],
    )
]

BAD_OFFENSE_TROOPS = [
    EffectsGroup(
        description="BAD_OFFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Ship & (DesignHasPart(name="GT_TROOP_POD") | DesignHasPart(name="GT_TROOP_POD_2")),
        stackinggroup="BAD_OFFENSIVE_TROOPS_STACK",
        accountinglabel="BAD_OFFENSIVE_TROOPS_LABEL",
        effects=[
            SetCapacity(partname="GT_TROOP_POD", value=PartCapacity(name="GT_TROOP_POD") * 0.5),
            SetCapacity(partname="GT_TROOP_POD_2", value=PartCapacity(name="GT_TROOP_POD_2") * 0.5),
        ],
    )
]


GOOD_OFFENSE_TROOPS = [
    EffectsGroup(
        description="GOOD_OFFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Ship & (DesignHasPart(name="GT_TROOP_POD") | DesignHasPart(name="GT_TROOP_POD_2")),
        stackinggroup="GOOD_OFFENSIVE_TROOPS_STACK",
        accountinglabel="GOOD_OFFENSIVE_TROOPS_LABEL",
        effects=[
            SetCapacity(partname="GT_TROOP_POD", value=PartCapacity(name="GT_TROOP_POD") * 1.5),
            SetCapacity(partname="GT_TROOP_POD_2", value=PartCapacity(name="GT_TROOP_POD_2") * 1.5),
        ],
    )
]


GREAT_OFFENSE_TROOPS = [
    EffectsGroup(
        description="GREAT_OFFENSE_TROOPS_DESC",
        scope=IsSource,
        activation=Ship & (DesignHasPart(name="GT_TROOP_POD") | DesignHasPart(name="GT_TROOP_POD_2")),
        stackinggroup="GREAT_OFFENSIVE_TROOPS_STACK",
        accountinglabel="GREAT_OFFENSIVE_TROOPS_LABEL",
        effects=[
            SetCapacity(partname="GT_TROOP_POD", value=PartCapacity(name="GT_TROOP_POD") * 2),
            SetCapacity(partname="GT_TROOP_POD_2", value=PartCapacity(name="GT_TROOP_POD_2") * 2),
        ],
    )
]



# Autogenerated do not modify manually!
# This is a type-hinting python stub file, used by python IDEs to provide type hints. For more information
# about stub files, see https://www.python.org/dev/peps/pep-0484/#stub-files
# During execution, the actual module is made available via
# a C++ Boost-python process as part of the launch.
from collections.abc import Mapping, Sequence, Set
from enum import IntEnum
from typing import overload

from common.fo_typing import (
    BuildingId,
    BuildingName,
    EmpireId,
    FleetId,
    NamedTuple,
    ObjectId,
    PartName,
    PlanetId,
    PlayerId,
    ShipId,
    SpeciesName,
    SystemId,
    Turn,
)

class AccountingInfo:
    @property
    def causeType(self): ...
    @property
    def customLabel(self): ...
    @property
    def meterChange(self): ...
    @property
    def meterRunningTotal(self): ...
    @property
    def sourceID(self): ...
    @property
    def specificCause(self): ...

class GalaxySetupData:
    @property
    def age(self) -> galaxySetupOptionGeneric: ...
    @property
    def gameUID(self) -> str: ...
    @property
    def maxAIAggression(self) -> aggression: ...
    @property
    def monsterFrequency(self) -> galaxySetupOptionMonsterFreq: ...
    @property
    def nativeFrequency(self) -> galaxySetupOptionGeneric: ...
    @property
    def planetDensity(self) -> galaxySetupOptionGeneric: ...
    @property
    def seed(self) -> str: ...
    @property
    def shape(self) -> galaxyShape: ...
    @property
    def size(self) -> int: ...
    @property
    def specialsFrequency(self) -> galaxySetupOptionGeneric: ...
    @property
    def starlaneFrequency(self) -> galaxySetupOptionGeneric: ...

class GameRules:
    @property
    def empty(self) -> bool: ...
    def getDescription(self, string: str) -> str: ...
    def getDouble(self, string: str) -> float: ...
    def getInt(self, string: str) -> int: ...
    def getRulesAsStrings(self) -> Mapping[str, str]: ...
    def getString(self, string: str) -> str: ...
    def getToggle(self, string: str) -> bool: ...
    def ruleExists(self, string: str) -> bool: ...
    def ruleExistsWithType(self, string: str, rule_type: ruleType) -> bool: ...

class UnlockableItem:
    @property
    def name(self) -> str: ...
    @property
    def type(self) -> unlockableItemType: ...

class buildingType:
    @property
    def costTimeLocationInvariant(self) -> bool: ...
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    def canBeEnqueued(self, number1: int, number2: int) -> bool: ...
    def canBeProduced(self, number1: int, number2: int) -> bool: ...
    def captureResult(self, number1: int, number2: int, number3: int, boolean: bool) -> captureResult: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """
    def perTurnCost(self, number1: int, number2: int) -> float: ...
    def productionCost(self, number1: int, number2: int) -> float: ...
    def productionTime(self, number1: int, number2: int) -> int: ...

class diplomaticMessage(NamedTuple):
    recipient: PlayerId
    sender: PlayerId
    type: diplomaticMessageType

class diplomaticStatusUpdate:
    @property
    def empire1(self) -> EmpireId: ...
    @property
    def empire2(self) -> EmpireId: ...
    @property
    def status(self) -> diplomaticStatus: ...

class empire:
    @property
    def adoptedPolicies(self) -> Sequence[str]: ...
    @property
    def allShipDesigns(self) -> Set[int]: ...
    @property
    def availableBuildingTypes(self) -> Sequence[str]: ...
    @property
    def availablePolicies(self) -> Set[str]: ...
    @property
    def availableShipDesigns(self) -> Set[int]: ...
    @property
    def availableShipHulls(self) -> Sequence[str]: ...
    @property
    def availableShipParts(self) -> Sequence[str]: ...
    @property
    def availableTechs(self) -> Set[str]: ...
    @property
    def capitalID(self) -> PlanetId: ...
    @property
    def categoriesSlotPolicies(self) -> Mapping[str, Mapping[int, str]]: ...
    @property
    def colour(self) -> tuple: ...
    @property
    def eliminated(self) -> bool: ...
    @property
    def empireID(self) -> EmpireId: ...
    @property
    def emptyPolicySlots(self) -> Mapping[str, int]: ...
    @property
    def exploredSystemIDs(self): ...
    @property
    def fleetSupplyableSystemIDs(self) -> Set[int]: ...
    @property
    def lastTurnReceived(self) -> int: ...
    @property
    def name(self) -> str: ...
    @property
    def planetsWithAllocatedPP(self) -> Mapping[Set[int], float]: ...
    @property
    def planetsWithAvailablePP(self) -> Mapping[Set[int], float]: ...
    @property
    def planetsWithWastedPP(self) -> Set[Set[int]]: ...
    @property
    def playerName(self) -> str: ...
    @property
    def productionPoints(self) -> float: ...
    @property
    def productionQueue(self) -> productionQueue: ...
    @property
    def researchQueue(self) -> researchQueue: ...
    @property
    def resourceSupplyGroups(self) -> Set[Set[int]]: ...
    @property
    def supplyUnobstructedSystems(self) -> Set[int]: ...
    @property
    def systemSupplyRanges(self) -> Mapping[int, float]: ...
    @property
    def totalPolicySlots(self) -> Mapping[str, int]: ...
    @property
    def totalShipsOwned(self) -> int: ...
    @property
    def turnsPoliciesAdopted(self) -> Mapping[str, int]: ...
    @property
    def won(self) -> bool: ...
    def buildingTypeAvailable(self, string: str) -> bool: ...
    @overload
    def canBuild(self, build_type: buildType, string: str, number: int) -> bool: ...
    @overload
    def canBuild(self, build_type: buildType, number1: int, number2: int) -> bool: ...
    def canBuild(*args) -> bool: ...
    def getMeter(self, string: str) -> meter:
        """
        Returns the empire meter with the indicated name (string).
        """
    def getTechStatus(self, string: str) -> techStatus: ...
    def hasExploredSystem(self, number: int) -> bool: ...
    def obstructedStarlanes(self) -> Sequence[tuple[int, int]]: ...
    def policyAdopted(self, string: str) -> bool: ...
    def policyAvailable(self, string: str) -> bool: ...
    def policyPrereqsAndExclusionsOK(self, string: str) -> bool: ...
    def population(self) -> float: ...
    def preservedLaneTravel(self, number1: int, number2: int) -> bool: ...
    def productionCostAndTime(self, production_queue_element: productionQueueElement) -> object: ...
    def researchProgress(self, string: str) -> float: ...
    def resourceAvailable(self, resource_type: resourceType) -> float: ...
    def resourceProduction(self, resource_type: resourceType) -> float: ...
    def resourceStockpile(self, resource_type: resourceType) -> float: ...
    def shipDesignAvailable(self, number: int) -> bool: ...
    def slotPolicyAdoptedIn(self, string: str) -> int: ...
    def supplyProjections(self) -> Mapping[SystemId, int]:
        """
        Returns the (negative) number of jumps (int) away each known system ID (int) is from this empire's supply network. 0 in dicates systems that are fleet supplied. -1 indicates a system that is 1 jump away from a supplied system. -4 indicates a system that is 4 jumps from a supply connection.
        """
    def techResearched(self, string: str) -> bool: ...
    def turnPolicyAdopted(self, string: str) -> Turn: ...

class fieldType:
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """

class meter:
    @property
    def current(self) -> float: ...
    @property
    def initial(self) -> float: ...
    def dump(self) -> str:
        """
        Returns string with debug information.
        """

class policy:
    @property
    def category(self) -> str: ...
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def shortDescription(self) -> str: ...
    @overload
    def adoptionCost(self) -> float: ...
    @overload
    def adoptionCost(self, empire_object: empire) -> float: ...
    @overload
    def adoptionCost(self, number: int) -> float: ...
    def adoptionCost(*args) -> float: ...

class productionQueue:
    @property
    def allocatedPP(self) -> Mapping[Set[int], float]: ...
    @property
    def empireID(self) -> EmpireId: ...
    @property
    def empty(self) -> bool: ...
    @property
    def size(self) -> int: ...
    @property
    def totalSpent(self) -> float: ...
    def __getitem__(self, number: int) -> productionQueueElement: ...
    def __iter__(self) -> object: ...
    def __len__(self) -> int: ...

class productionQueueElement:
    @property
    def allocation(self) -> float: ...
    @property
    def allowedStockpile(self) -> bool: ...
    @property
    def blocksize(self) -> int: ...
    @property
    def buildType(self) -> buildType: ...
    @property
    def designID(self) -> int: ...
    @property
    def locationID(self) -> PlanetId: ...
    @property
    def name(self) -> str: ...
    @property
    def paused(self) -> bool: ...
    @property
    def progress(self) -> float: ...
    @property
    def remaining(self) -> int: ...
    @property
    def removed(self) -> bool: ...
    @property
    def turnsLeft(self) -> int: ...

class researchQueue:
    @property
    def empireID(self) -> EmpireId: ...
    @property
    def empty(self) -> bool: ...
    @property
    def size(self) -> int: ...
    @property
    def totalSpent(self) -> float: ...
    def __contains__(self, research_queue_element: researchQueueElement) -> bool: ...
    def __getitem__(self, number: int) -> researchQueueElement: ...
    def __iter__(self) -> object: ...
    def __len__(self) -> int: ...
    def inQueue(self, string: str) -> bool: ...

class researchQueueElement:
    @property
    def allocation(self) -> float: ...
    @property
    def tech(self) -> str: ...
    @property
    def turnsLeft(self) -> int: ...

class shipDesign:
    @property
    def attack(self) -> float: ...
    @property
    def attackStats(self) -> Sequence[int]: ...
    @property
    def canColonize(self) -> bool: ...
    @property
    def canInvade(self) -> bool: ...
    @property
    def colonyCapacity(self) -> float: ...
    @property
    def costTimeLocationInvariant(self) -> bool: ...
    @property
    def defense(self) -> float: ...
    @property
    def description(self) -> str: ...
    @property
    def designedOnTurn(self) -> Turn: ...
    @property
    def detection(self) -> float: ...
    @property
    def fuel(self) -> float: ...
    @property
    def hasDirectWeapons(self) -> bool: ...
    @property
    def hasFighters(self) -> bool: ...
    @property
    def hull(self) -> str: ...
    @property
    def id(self) -> int: ...
    @property
    def industryGeneration(self) -> float: ...
    @property
    def influenceGeneration(self) -> float: ...
    @property
    def isArmed(self) -> bool: ...
    @property
    def isMonster(self) -> bool: ...
    @property
    def name(self) -> str: ...
    @property
    def parts(self) -> Sequence[PartName]: ...
    @property
    def researchGeneration(self) -> float: ...
    @property
    def shields(self) -> float: ...
    @property
    def ship_hull(self) -> shipHull: ...
    @property
    def speed(self) -> float: ...
    @property
    def stealth(self) -> float: ...
    @property
    def structure(self) -> float: ...
    @property
    def troopCapacity(self) -> float: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """
    def perTurnCost(self, number1: int, number2: int) -> float: ...
    def productionCost(self, number1: int, number2: int) -> float: ...
    def productionLocationForEmpire(self, number1: int, number2: int) -> bool: ...
    def productionTime(self, number1: int, number2: int) -> int: ...

class shipHull:
    @property
    def costTimeLocationInvariant(self) -> bool: ...
    @property
    def fuel(self) -> float: ...
    @property
    def name(self) -> str: ...
    @property
    def numSlots(self) -> int: ...
    @property
    def slots(self) -> Sequence[shipSlotType]: ...
    @property
    def speed(self) -> float: ...
    @property
    def stealth(self) -> float: ...
    @property
    def structure(self) -> float: ...
    def hasTag(self, string: str) -> bool: ...
    def numSlotsOfSlotType(self, ship_slot_type: shipSlotType) -> int: ...
    def productionCost(self, number1: int, number2: int, number3: int) -> float: ...
    def productionLocation(self, number: int) -> bool:
        """
        Returns the result of Location condition (bool) in passed location_id (int)
        """
    def productionTime(self, number1: int, number2: int, number3: int) -> int: ...

class shipPart:
    @property
    def capacity(self) -> float: ...
    @property
    def costTimeLocationInvariant(self) -> bool: ...
    @property
    def mountableSlotTypes(self) -> Sequence[shipSlotType]: ...
    @property
    def name(self) -> str: ...
    @property
    def partClass(self) -> shipPartClass: ...
    @property
    def secondaryStat(self) -> float: ...
    def canMountInSlotType(self, ship_slot_type: shipSlotType) -> bool: ...
    def hasTag(self, string: str) -> bool: ...
    def productionCost(self, number1: int, number2: int, number3: int) -> float: ...
    def productionLocation(self, number: int) -> bool:
        """
        Returns the result of Location condition (bool) in passed location_id (int)
        """
    def productionTime(self, number1: int, number2: int, number3: int) -> int: ...

class special:
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def spawnlimit(self) -> int: ...
    @property
    def spawnrate(self) -> float: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """
    def initialCapacity(self, number: int) -> float: ...

class species:
    @property
    def canColonize(self) -> bool: ...
    @property
    def canProduceShips(self) -> bool: ...
    @property
    def description(self) -> str: ...
    @property
    def dislikes(self) -> Sequence[str]: ...
    @property
    def foci(self) -> Sequence[str]: ...
    @property
    def homeworlds(self) -> Set[int]: ...
    @property
    def likes(self) -> Sequence[str]: ...
    @property
    def name(self) -> str: ...
    @property
    def native(self) -> bool: ...
    @property
    def preferredFocus(self) -> str: ...
    @property
    def spawnlimit(self) -> int: ...
    @property
    def spawnrate(self) -> float: ...
    @property
    def tags(self) -> Sequence[str]: ...
    def dump(self, number: int) -> str:
        """
        Returns string with debug information, use '0' as argument.
        """
    def getPlanetEnvironment(self, planet_type: planetType) -> planetEnvironment: ...

class tech:
    @property
    def category(self) -> str: ...
    @property
    def description(self) -> str: ...
    @property
    def name(self) -> str: ...
    @property
    def prerequisites(self) -> Sequence[str]: ...
    @property
    def shortDescription(self) -> str: ...
    @property
    def unlockedItems(self) -> Sequence[UnlockableItem]: ...
    @property
    def unlockedTechs(self) -> Sequence[str]: ...
    def perTurnCost(self, number: int) -> float: ...
    def recursivePrerequisites(self, number: int) -> Sequence[str]: ...
    def researchCost(self, number: int) -> float: ...
    def researchTime(self, number: int) -> int: ...

class universe:
    @property
    def buildingIDs(self) -> Sequence[BuildingId]: ...
    @property
    def effectAccounting(self) -> Mapping[int, Mapping[meterType, Sequence[AccountingInfo]]]: ...
    @property
    def fieldIDs(self) -> Sequence[int]: ...
    @property
    def fleetIDs(self) -> Sequence[FleetId]: ...
    @property
    def planetIDs(self) -> Sequence[PlanetId]: ...
    @property
    def shipIDs(self) -> Sequence[ShipId]: ...
    @property
    def systemIDs(self) -> Sequence[SystemId]: ...
    def destroyedObjectIDs(self, number: int) -> Set[int]: ...
    def dump(self) -> None: ...
    def getBuilding(self, number: int) -> building: ...
    def getField(self, number: int) -> field: ...
    def getFleet(self, number: int) -> fleet: ...
    def getGenericShipDesign(self, string: str) -> shipDesign:
        """
        Returns the ship design (ShipDesign) with the indicated name (string).
        """
    def getImmediateNeighbors(self, number1: int, number2: int) -> Sequence[int]: ...
    def getObject(self, number: int) -> universeObject: ...
    def getPlanet(self, number: int) -> planet: ...
    def getShip(self, number: int) -> ship: ...
    def getSystem(self, number: int) -> system: ...
    def getSystemNeighborsMap(self, number1: int, number2: int) -> Mapping[int, float]: ...
    def getVisibility(self, number1: int, number2: int) -> visibility: ...
    def getVisibilityTurnsMap(self, number1: int, number2: int) -> Mapping[visibility, int]: ...
    def jumpDistance(self, number1: int, number2: int) -> int:
        """
        If two system ids are passed or both objects are within a system, return the jump distance between the two systems. If one object (e.g. a fleet) is on a starlane, then calculate the jump distance from both ends of the starlane to the target system and return the smaller one.
        """
    def leastJumpsPath(self, number1: int, number2: int, number3: int) -> Sequence[int]: ...
    def linearDistance(self, number1: int, number2: int) -> float: ...
    def shortestNonHostilePath(self, number1: int, number2: int, number3: int) -> Sequence[int]:
        """
        Shortest sequence of System ids and distance from System (number1) to System (number2) with no hostile Fleets as determined by visibility of Empire (number3).  (number3) must be a valid empire.
        """
    def shortestPath(self, number1: int, number2: int, number3: int) -> Sequence[int]: ...
    def shortestPathDistance(self, number1: int, number2: int) -> float: ...
    def statRecords(self) -> Mapping[str, Mapping[int, Mapping[int, float]]]:
        """
        Empire statistics recorded by the server each turn. Indexed first by staistic name (string), then by empire id (int), then by turn number (int), pointing to the statisic value (double).
        """
    def systemHasStarlane(self, number1: int, number2: int) -> bool: ...
    def systemsConnected(self, number1: int, number2: int, number3: int) -> bool: ...
    def updateMeterEstimates(self, obj: object) -> None: ...

class universeObject:
    @property
    def ageInTurns(self): ...
    @property
    def containedObjects(self): ...
    @property
    def containerObject(self): ...
    @property
    def creationTurn(self): ...
    @property
    def id(self) -> ObjectId: ...
    @property
    def meters(self): ...
    @property
    def name(self) -> str: ...
    @property
    def owner(self) -> EmpireId: ...
    @property
    def specials(self) -> SpeciesName: ...
    @property
    def systemID(self) -> SystemId: ...
    @property
    def tags(self): ...
    @property
    def unowned(self): ...
    @property
    def x(self): ...
    @property
    def y(self): ...
    def containedBy(self, number: int) -> bool: ...
    def contains(self, number: int) -> bool: ...
    def currentMeterValue(self, meter_type: meterType) -> float: ...
    def dump(self) -> str:
        """
        Returns string with debug information.
        """
    def getMeter(self, meter_type: meterType) -> meter: ...
    def hasSpecial(self, obj: object) -> bool: ...
    def hasTag(self, string: str) -> bool: ...
    def initialMeterValue(self, meter_type: meterType) -> float: ...
    def ownedBy(self, number: int) -> bool: ...
    def specialAddedOnTurn(self, obj: object) -> int: ...

class building(universeObject):
    @property
    def buildingTypeName(self) -> BuildingName: ...
    @property
    def orderedScrapped(self) -> bool: ...
    @property
    def planetID(self) -> PlanetId: ...
    @property
    def producedByEmpireID(self) -> EmpireId: ...

class field(universeObject):
    @property
    def fieldTypeName(self) -> str: ...
    @overload
    def inField(self, base_object: universeObject) -> bool: ...
    @overload
    def inField(self, floating_number1: float, floating_number2: float) -> bool: ...
    def inField(*args) -> bool: ...

class fleet(universeObject):
    @property
    def aggression(self) -> fleetAggression: ...
    @property
    def aggressive(self) -> bool: ...
    @property
    def canChangeDirectionEnRoute(self) -> bool: ...
    @property
    def empty(self) -> bool: ...
    @property
    def finalDestinationID(self) -> int: ...
    @property
    def fuel(self) -> float: ...
    @property
    def hasArmedShips(self) -> bool: ...
    @property
    def hasColonyShips(self) -> bool: ...
    @property
    def hasFighterShips(self) -> bool: ...
    @property
    def hasMonsters(self) -> bool: ...
    @property
    def hasOutpostShips(self) -> bool: ...
    @property
    def hasTroopShips(self) -> bool: ...
    @property
    def maxFuel(self) -> float: ...
    @property
    def nextSystemID(self) -> int: ...
    @property
    def numShips(self) -> int: ...
    @property
    def obstructive(self) -> bool: ...
    @property
    def previousSystemID(self) -> int: ...
    @property
    def route(self) -> Sequence[int]: ...
    @property
    def shipIDs(self) -> Sequence[ShipId]: ...
    @property
    def speed(self) -> float: ...

class planet(universeObject):
    @property
    def availableFoci(self) -> Sequence[str]: ...
    @property
    def buildingIDs(self) -> Sequence[BuildingId]: ...
    @property
    def clockwiseNextPlanetType(self) -> planetType: ...
    @property
    def counterClockwiseNextPlanetType(self) -> planetType: ...
    @property
    def distanceFromOriginalType(self) -> int: ...
    @property
    def focus(self) -> str: ...
    @property
    def habitableSize(self) -> int: ...
    @property
    def lastColonizedByEmpire(self) -> int: ...
    @property
    def lastInvadedByEmpire(self) -> int: ...
    @property
    def lastTurnAttackedByShip(self) -> int: ...
    @property
    def lastTurnColonized(self) -> int: ...
    @property
    def lastTurnConquered(self) -> int: ...
    @property
    def nextLargerPlanetSize(self) -> planetSize: ...
    @property
    def nextSmallerPlanetSize(self) -> planetSize: ...
    @property
    def originalType(self) -> planetType: ...
    @property
    def ownerBeforeLastConquered(self) -> int: ...
    @property
    def size(self) -> planetSize: ...
    @property
    def speciesName(self) -> SpeciesName: ...
    @property
    def turnsSinceFocusChange(self) -> int: ...
    @property
    def type(self) -> planetType: ...
    def OrbitalPositionOnTurn(self, number: int) -> float: ...
    def environmentForSpecies(self, string: str) -> planetEnvironment: ...
    def nextBetterPlanetTypeForSpecies(self, obj: object, string: str) -> planetType: ...

class ship(universeObject):
    @property
    def arrivedOnTurn(self) -> int: ...
    @property
    def canBombard(self) -> bool: ...
    @property
    def canColonize(self) -> bool: ...
    @property
    def canInvade(self) -> bool: ...
    @property
    def colonyCapacity(self) -> float: ...
    @property
    def design(self) -> shipDesign: ...
    @property
    def designID(self) -> int: ...
    @property
    def fleetID(self) -> int: ...
    @property
    def hasFighters(self) -> bool: ...
    @property
    def isArmed(self) -> bool: ...
    @property
    def isMonster(self) -> bool: ...
    @property
    def lastResuppliedOnTurn(self) -> int: ...
    @property
    def lastTurnActiveInCombat(self) -> int: ...
    @property
    def orderedColonizePlanet(self) -> int: ...
    @property
    def orderedInvadePlanet(self) -> int: ...
    @property
    def orderedScrapped(self) -> bool: ...
    @property
    def partMeters(self) -> Mapping[tuple[meterType, meterType], meter]: ...
    @property
    def producedByEmpireID(self) -> EmpireId: ...
    @property
    def speciesName(self) -> SpeciesName: ...
    @property
    def speed(self) -> float: ...
    @property
    def troopCapacity(self) -> float: ...
    def currentPartMeterValue(self, meter_type: meterType, string: str) -> float: ...
    def initialPartMeterValue(self, meter_type: meterType, string: str) -> float: ...

class system(universeObject):
    @property
    def buildingIDs(self) -> Sequence[BuildingId]: ...
    @property
    def fieldIDs(self) -> Sequence[int]: ...
    @property
    def fleetIDs(self) -> Sequence[FleetId]: ...
    @property
    def lastTurnBattleHere(self) -> int: ...
    @property
    def numStarlanes(self) -> int: ...
    @property
    def planetIDs(self) -> Sequence[PlanetId]: ...
    @property
    def shipIDs(self) -> Sequence[ShipId]: ...
    @property
    def starType(self) -> starType: ...
    @property
    def starlanes(self) -> Sequence[int]: ...
    @property
    def starlanesWormholes(self) -> Sequence[int]: ...
    def HasStarlaneToSystemID(self, number: int) -> bool:
        """
        true if the passed in ID (int) is the ID of a system this system has a starlane connection with
        """

class aggression(IntEnum):
    invalid = -1
    beginner = 0
    turtle = 1
    cautious = 2
    typical = 3
    aggressive = 4
    maniacal = 5

class buildType(IntEnum):
    INVALID_BUILD_TYPE = -1
    BT_NOT_BUILDING = 0
    BT_BUILDING = 1
    BT_SHIP = 2
    BT_PROJECT = 3
    BT_STOCKPILE = 4
    NUM_BUILD_TYPES = 5

class captureResult(IntEnum):
    capture = 0
    destroy = 1
    retain = 2

class diplomaticMessageType(IntEnum):
    noMessage = -1
    warDeclaration = 0
    peaceProposal = 1
    acceptPeaceProposal = 2
    alliesProposal = 3
    acceptAlliesProposal = 4
    endAllies = 5
    cancelProposal = 6
    rejectProposal = 7

class diplomaticStatus(IntEnum):
    war = 0
    peace = 1
    allied = 2

class effectsCauseType(IntEnum):
    invalid = -1
    unknown = 0
    inherent = 1
    tech = 2
    building = 3
    field = 4
    special = 5
    species = 6
    shipPart = 7
    shipHull = 8
    policy = 9

class fleetAggression(IntEnum):
    passive = 0
    defensive = 1
    obstructive = 2
    aggressive = 3

class galaxySetupOptionGeneric(IntEnum):
    invalid = -1
    none = 0
    low = 1
    medium = 2
    high = 3
    random = 4

class galaxySetupOptionMonsterFreq(IntEnum):
    invalid = -1
    none = 0
    extremelyLow = 1
    veryLow = 2
    low = 3
    medium = 4
    high = 5
    veryHigh = 6
    extremelyHigh = 7
    random = 8

class galaxyShape(IntEnum):
    invalid = -1
    spiral2 = 0
    spiral3 = 1
    spiral4 = 2
    cluster = 3
    elliptical = 4
    disc = 5
    box = 6
    irregular = 7
    ring = 8
    random = 9

class meterType(IntEnum):
    targetPopulation = 0
    targetIndustry = 1
    targetResearch = 2
    targetInfluence = 3
    targetConstruction = 4
    targetHappiness = 5
    maxCapacity = 6
    maxSecondaryStat = 7
    maxFuel = 8
    maxShield = 9
    maxStructure = 10
    maxDefense = 11
    maxSupply = 12
    maxStockpile = 13
    maxTroops = 14
    population = 15
    industry = 16
    research = 17
    influence = 18
    construction = 19
    happiness = 20
    capacity = 21
    secondaryStat = 22
    fuel = 23
    shield = 24
    structure = 25
    defense = 26
    supply = 27
    stockpile = 28
    troops = 29
    rebels = 30
    size = 31
    stealth = 32
    detection = 33
    speed = 34

class planetEnvironment(IntEnum):
    uninhabitable = 0
    hostile = 1
    poor = 2
    adequate = 3
    good = 4

class planetSize(IntEnum):
    unknown = -1
    noWorld = 0
    tiny = 1
    small = 2
    medium = 3
    large = 4
    huge = 5
    asteroids = 6
    gasGiant = 7

class planetType(IntEnum):
    unknown = -1
    swamp = 0
    toxic = 1
    inferno = 2
    radiated = 3
    barren = 4
    tundra = 5
    desert = 6
    terran = 7
    ocean = 8
    asteroids = 9
    gasGiant = 10

class resourceType(IntEnum):
    industry = 0
    influence = 1
    research = 2
    stockpile = 3

class roleType(IntEnum):
    host = 0
    clientTypeModerator = 1
    clientTypePlayer = 2
    clientTypeObserver = 3
    galaxySetup = 4

class ruleType(IntEnum):
    invalid = -1
    toggle = 0
    int = 1
    double = 2
    string = 3

class shipPartClass(IntEnum):
    shortRange = 0
    fighterBay = 1
    fighterHangar = 2
    shields = 3
    armour = 4
    troops = 5
    detection = 6
    stealth = 7
    fuel = 8
    colony = 9
    speed = 10
    general = 11
    bombard = 12
    industry = 13
    research = 14
    influence = 15
    productionLocation = 16

class shipSlotType(IntEnum):
    external = 0
    internal = 1
    core = 2

class starType(IntEnum):
    unknown = -1
    blue = 0
    white = 1
    yellow = 2
    orange = 3
    red = 4
    neutron = 5
    blackHole = 6
    noStar = 7

class techStatus(IntEnum):
    unresearchable = 0
    partiallyUnlocked = 1
    researchable = 2
    complete = 3

class unlockableItemType(IntEnum):
    invalid = -1
    building = 0
    shipPart = 1
    shipHull = 2
    shipDesign = 3
    tech = 4
    policy = 5

class visibility(IntEnum):
    invalid = -1
    none = 0
    basic = 1
    partial = 2
    full = 3

def allEmpireIDs() -> Sequence[EmpireId]:
    """
    Returns an object (intVec) that contains the empire IDs of all empires in the game.
    """

def allPlayerIDs() -> Sequence[int]:
    """
    Returns an object (intVec) that contains the player IDs of all players in the game.
    """

def appendFleetMoveOrder(number1: int, number2: int) -> int:
    """
    Orders the fleet with indicated fleetID (int) to append the system with the indicated destinationID (int) to its possibly already enqueued route. Returns 1 (int) on success or 0 (int) on failure due to not finding the indicated fleet or system.
    """

def currentTurn() -> Turn:
    """
    Returns the current game turn (int).
    """

def empireID() -> EmpireId:
    """
    Returns the empire ID (int) of this AI player's empire.
    """

def empirePlayerID(number: int) -> PlayerId:
    """
    Returns the player ID (int) of the player who is controlling the empire with the indicated empireID (int).
    """

def getAIDir() -> str: ...
def getBuildingType(string: str) -> buildingType:
    """
    Returns the building type (BuildingType) with the indicated name (string).
    """

def getDiplomaticStatus(number1: int, number2: int) -> diplomaticStatus:
    """
    Returns the diplomatic status between two empires
    """

@overload
def getEmpire() -> empire: ...
@overload
def getEmpire(number: int) -> empire: ...
def getEmpire(*args) -> empire:
    """
    Returns the empire object (Empire) of this AI player
    Returns the empire object (Empire) with the specified empire ID (int)
    """

def getFieldType(string: str) -> fieldType: ...
def getGalaxySetupData() -> GalaxySetupData: ...
def getGameRules() -> GameRules:
    """
    Returns the game rules manager, which can be used to look up the names (string) of rules are defined with what type (boolean / toggle, int, double, string), and what values the rules have in the current game.
    """

def getNamedInt(string: str) -> int:
    """
    Returns the named integer value of the scripted constant with name (string). If no such named constant exists, returns 0.
    """

def getNamedReal(string: str) -> float:
    """
    Returns the named real value of the scripted constant with name (string). If no such named constant exists, returns 0.0.
    """

def getOptionsDBOptionBool(string: str) -> bool:
    """
    Returns the bool value of option in OptionsDB or None if the option does not exist.
    """

def getOptionsDBOptionDouble(string: str) -> float:
    """
    Returns the double value of option in OptionsDB or None if the option does not exist.
    """

def getOptionsDBOptionInt(string: str) -> int:
    """
    Returns the integer value of option in OptionsDB or None if the option does not exist.
    """

def getOptionsDBOptionStr(string: str) -> str:
    """
    Returns the string value of option in OptionsDB or None if the option does not exist.
    """

def getPolicy(string: str) -> policy:
    """
    Returns the policy (Policy) with the indicated name (string).
    """

def getPredefinedShipDesign(string: str) -> shipDesign:
    """
    Returns the ship design (ShipDesign) with the indicated name (string).
    """

def getSaveStateString() -> str:
    """
    Returns the previously-saved state string (string). Can be used to retrieve the last-set save state string at any time, although this string is also passed to the resumeLoadedGame(savedStateString) Python function when a game is loaded, so this function isn't necessary to use if resumeLoadedGame stores the passed string.
    """

def getShipDesign(number: int) -> shipDesign:
    """
    Returns the ship design (ShipDesign) with the indicated id number (int).
    """

def getShipHull(string: str) -> shipHull:
    """
    Returns the ship hull with the indicated name (string).
    """

def getShipPart(string: str) -> shipPart:
    """
    Returns the ShipPart with the indicated name (string).
    """

def getSpecial(string: str) -> special:
    """
    Returns the special (Special) with the indicated name (string).
    """

def getSpecies(string: str) -> species:
    """
    Returns the species (Species) with the indicated name (string).
    """

def getTech(string: str) -> tech:
    """
    Returns the tech (Tech) with the indicated name (string).
    """

def getTechCategories() -> Sequence[str]:
    """
    Returns the names of all tech categories (StringVec).
    """

def getUniverse() -> universe:
    """
    Returns the universe object (Universe)
    """

def getUserConfigDir() -> str:
    """
    Returns path to directory where FreeOrion stores user specific configuration.
    """

def getUserDataDir() -> str:
    """
    Returns path to directory where FreeOrion stores user specific data (saves, etc.).
    """

def initMeterEstimatesDiscrepancies() -> None:
    """
    For all objects and max / target meters, determines discrepancies between actual meters and what the known universe should produce. This is used later when updating meter estimates to incorporate those discrepancies.
    """

def isEnqueuableBuilding(string: str, number: int) -> bool:
    """
    Returns true if the specified building type (string) can be enqueued by this client's empire at the specified production location (int). Being enqueuable means that the item can be added to the queue, but does not mean that the item will be allocated production points once  it is added
    """

def isEnqueuableShip(number1: int, number2: int) -> bool:
    """
    Returns true if the specified ship design (int) can be enqueued by this client's empire at the specified production location (int). Enqueued ships should always also be producible, and thus able to be allocated production points once enqueued, if any are available at the production location.
    """

def isProducibleBuilding(string: str, number: int) -> bool:
    """
    Returns true if the specified building type (string) can be produced by this client's empire at the specified production location (int). Being producible means that if the item is on the production queue, it can be allocated production points that are available at its location. Being producible does not mean that the building type can be added to the queue.
    """

def issueAdoptPolicyOrder(string1: str, string2: str, number: int) -> int:
    """
    Orders the policy with name policyName (string) to be adopted by the empire in the category categoryName (string) and slot slot (int). Returns 1 (int) on success or 0 (int) on failure if the indicated policy can't be found. Will return 1 (int) but do nothing if the indicated policy can't be enqueued by this player's empire in the requested category and slot.
    """

def issueAggressionOrder(number: int, boolean: bool) -> int: ...
def issueAllowStockpileProductionOrder(number: int, boolean: bool) -> int:
    """
    Orders the item on the production queue at index queueIndex (int) to be enabled (or disabled) to use the imperial stockpile. Returns 1 (int) on success or 0 (int) on failure if the queue index is less than 0 or greater than the largest indexed item on the queue.
    """

def issueBombardOrder(number1: int, number2: int) -> int: ...
def issueChangeFocusOrder(number: int, string: str) -> int:
    """
    Orders the planet with ID planetID (int) to use focus setting focus (string). Returns 1 (int) on success or 0 (int) on failure if the planet can't be found or isn't owned by this player, or if the specified focus is not valid on the planet.
    """

def issueChangeProductionQuantityOrder(number1: int, number2: int, number3: int) -> int: ...
def issueColonizeOrder(number1: int, number2: int) -> int:
    """
    Orders the ship with ID shipID (int) to colonize the planet with ID planetID (int). Returns 1 (int) on success or 0 (int) on failure due to not finding the indicated ship or planet, this client's player not owning the indicated ship, the planet already being colonized, or the planet and ship not being in the same system.
    """

def issueCreateShipDesignOrder(
    string1: str,
    string2: str,
    string3: str,
    item_list: list,
    string4: str,
    string5: str,
    boolean: bool,
) -> int:
    """
    Orders the creation of a new ship design with the name (string), description (string), hull (string), parts vector partsVec (StringVec), graphic (string) and model (string). model should be left as an empty string as of this writing. There is currently no easy way to find the id of the new design, though the client's empire should have the new design after this order is issued successfully. Returns 1 (int) on success or 0 (int) on failure if any of the name, description, hull or graphic are empty strings, if the design is invalid (due to not following number and type of slot requirements for the hull) or if creating the design fails for some reason.
    """

def issueDeadoptPolicyOrder(string: str) -> int:
    """
    Orders the policy with name policyName (string) to be de-adopted by the empire. Returns 1 (int) on success or 0 (int) on failure if the indicated policy was not already adopted.
    """

def issueDequeueProductionOrder(number: int) -> int:
    """
    Orders the item on the production queue at index queueIndex (int) to be removed form the production queue. Returns 1 (int) on success or 0 (int) on failure if the queue index is less than 0 or greater than the largest indexed item on the queue.
    """

def issueDequeueTechOrder(string: str) -> int:
    """
    Orders the tech with name techName (string) to be removed from the queue. Returns 1 (int) on success or 0 (int) on failure if the indicated tech can't be found. Will return 1 (int) but do nothing if the indicated tech isn't on this player's empire's tech queue.
    """

def issueEnqueueBuildingProductionOrder(string: str, number: int) -> int:
    """
    Orders the building with name (string) to be added to the production queue at the location of the planet with id locationID. Returns 1 (int) on success or 0 (int) on failure if there is no such building or it is not available to this player's empire, or if the building can't be produced at the specified location.
    """

def issueEnqueueShipProductionOrder(number1: int, number2: int) -> int:
    """
    Orders the ship design with ID designID (int) to be added to the production queue at the location of the planet with id locationID (int). Returns 1 (int) on success or 0 (int) on failure there is no such ship design or it not available to this player's empire, or if the design can't be produced at the specified location.
    """

def issueEnqueueTechOrder(string: str, number: int) -> int:
    """
    Orders the tech with name techName (string) to be added to the tech queue at position (int) on the queue. Returns 1 (int) on success or 0 (int) on failure if the indicated tech can't be found. Will return 1 (int) but do nothing if the indicated tech can't be enqueued by this player's empire.
    """

def issueFleetMoveOrder(number1: int, number2: int) -> int:
    """
    Orders the fleet with indicated fleetID (int) to move to the system with the indicated destinationID (int). Returns 1 (int) on success or 0 (int) on failure due to not finding the indicated fleet or system.
    """

def issueFleetTransferOrder(number1: int, number2: int) -> int:
    """
    Orders the ship with ID shipID (int) to be transferred to the fleet with ID newFleetID. Returns 1 (int) on success, or 0 (int) on failure due to not finding the fleet or ship, or the client's empire not owning either, or the two not being in the same system (or either not being in a system) or the ship already being in the fleet.
    """

def issueGiveObjectToEmpireOrder(number1: int, number2: int) -> int: ...
def issueInvadeOrder(number1: int, number2: int) -> int: ...
def issueNewFleetOrder(string: str, number: int) -> int:
    """
    Orders a new fleet to be created with the indicated name (string) and containing the indicated shipIDs (IntVec). The ships must be located in the same system and must all be owned by this player. Returns the new fleets id (int) on success or 0 (int) on failure due to one of the noted conditions not being met.
    """

def issuePauseProductionOrder(number: int, boolean: bool) -> int:
    """
    Orders the item on the production queue at index queueIndex (int) to be paused (or unpaused). Returns 1 (int) on success or 0 (int) on failure if the queue index is less than 0 or greater than the largest indexed item on the queue.
    """

def issueRenameOrder(number: int, string: str) -> int:
    """
    Orders the renaming of the object with indicated objectID (int) to the new indicated name (string). Returns 1 (int) on success or 0 (int) on failure due to this AI player not being able to rename the indicated object (which this player must fully own, and which must be a fleet, ship or planet).
    """

def issueRequeueProductionOrder(number1: int, number2: int) -> int:
    """
    Orders the item on the production queue at index oldQueueIndex (int) to be moved to index newQueueIndex (int). Returns 1 (int) on success or 0 (int) on failure if the old and new queue indices are equal, if either queue index is less than 0 or greater than the largest indexed item on the queue.
    """

def issueScrapOrder(number: int) -> int:
    """
    Orders the ship or building with the indicated objectID (int) to be scrapped. Returns 1 (int) on success or 0 (int) on failure due to not finding a ship or building with the indicated ID, or if the indicated ship or building is not owned by this AI client's empire.
    """

def namedIntDefined(string: str) -> bool:
    """
    Returns true/false (boolean) whether there is a defined int-valued scripted constant with name (string).
    """

def namedRealDefined(string: str) -> bool:
    """
    Returns true/false (boolean) whether there is a defined double-valued scripted constant with name (string).
    """

def playerEmpireID(number: int) -> int:
    """
    Returns the empire ID (int) of the player with the specified player ID (int).
    """

def playerID() -> int:
    """
    Returns the integer id of this AI player.
    """

def playerIsAI(number: int) -> bool:
    """
    Returns True (boolean) if the player with the indicated playerID (int) is controlled by an AI and false (boolean) otherwise.
    """

def playerIsHost(number: int) -> bool:
    """
    Returns True (boolean) if the player with the indicated playerID (int) is the host player for the game and false (boolean) otherwise.
    """

@overload
def playerName() -> str: ...
@overload
def playerName(number: int) -> str: ...
def playerName(*args) -> str:
    """
    Returns the name (string) of this AI player.
    Returns the name (string) of the player with the indicated playerID (int).
    """

def policies() -> Sequence[str]:
    """
    Returns the names of all policies (StringVec).
    """

def policiesInCategory(string: str) -> Sequence[str]:
    """
    Returns the names of all policies (StringVec) in the indicated policy category name (string).
    """

def policyCategories() -> Sequence[str]:
    """
    Returns the names of all policy categories (StringVec).
    """

def sendChatMessage(number: int, string: str) -> None:
    """
    Sends the indicated message (string) to the player with the indicated recipientID (int) or to all players if recipientID is -1.
    """

def sendDiplomaticMessage(diplomatic_message: diplomaticMessage) -> None: ...
def setSaveStateString(string: str) -> None:
    """
    Sets the save state string (string). This is a persistant storage space for the AI script to retain state information when the game is saved and reloaded. Any AI state information to be saved should be stored in a single string (likely using Python's pickle module) and stored using this function when the prepareForSave() Python function is called.
    """

def techs() -> Sequence[str]:
    """
    Returns the names of all techs (StringVec).
    """

def techsInCategory(string: str) -> Sequence[str]:
    """
    Returns the names of all techs (StringVec) in the indicated tech category name (string).
    """

def updateMeterEstimates(boolean: bool) -> None: ...
def updateProductionQueue() -> None: ...
def updateResearchQueue() -> None: ...
def updateResourcePools() -> None: ...
def userString(string: str) -> str: ...
def userStringExists(string: str) -> bool: ...
def userStringList(string: str) -> list: ...
def validShipDesign(string: str, item_list: list) -> bool:
    """
    Returns true (boolean) if the passed hull (string) and parts (list of string) make up a valid ship design, and false (boolean) otherwise. Valid ship designs don't have any parts in slots that can't accept that type of part, and contain only hulls and parts that exist (and may also need to contain the correct number of parts - this needs to be verified).
    """



# usually each "level" of fuel allows two jumps fuel
# (i.e. GREAT_FUEL is similar to GREAT_SUPPLY)
from focs._effects import EffectsGroup, IsSource, SetMaxFuel, Ship, Value
from macros.priorities import TARGET_AFTER_SCALING_PRIORITY


def _fuel(label: str, fuel_change: float):
    return EffectsGroup(
        description=f"{label}_FUEL_DESC",
        scope=IsSource & Ship,
        accountinglabel=f"{label}_FUEL_LABEL",
        priority=TARGET_AFTER_SCALING_PRIORITY,
        effects=SetMaxFuel(value=fuel_change),
    )


NO_FUEL = [
    _fuel("NO", Value + 0),
]
BAD_FUEL = [
    _fuel("BAD", Value - 0.5),
]
GOOD_FUEL = [
    _fuel("GOOD", Value + 0.5),
]
AVERAGE_FUEL = []
GREAT_FUEL = [
    _fuel("GREAT", Value + 1),
]
ULTIMATE_FUEL = [
    _fuel("ULTIMATE", Value + 1.5),
]



from operator import attrgetter

from stub_generator.interface_inspector import EnumInfo


def _handle_enum(info: EnumInfo):
    def attr_sort_key(pair):
        """
        Make key to sort attributes.

        We sort enum items by is value and after that by name.
        Enum item could have aliases with the same value and different names.
        """
        name, val = pair
        return val, name

    pairs = sorted(info.attributes.items(), key=attr_sort_key)
    result = [
        "class %s(IntEnum):" % info.name,
    ]

    for text, value in pairs:
        result.append(f"    {text} = {value}")
    return "\n".join(result)


def generate_enums(enums: list[EnumInfo]):
    for enum in sorted(enums, key=attrgetter("name")):
        yield _handle_enum(enum)



from focs._effects import (
    EffectsGroup,
    HasSpecies,
    OwnedBy,
    SetTargetConstruction,
    SetTargetPopulation,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER
from macros.priorities import TARGET_POPULATION_AFTER_SCALING_PRIORITY

Tech(
    name="CON_NDIM_STRC",
    description="CON_NDIM_STRC_DESC",
    short_description="POPULATION_SHORT_DESC",
    category="CONSTRUCTION_CATEGORY",
    researchcost=630 * TECH_COST_MULTIPLIER,
    researchturns=9,
    tags=["PEDIA_CONSTRUCTION_CATEGORY"],
    prerequisites=["CON_FRC_ENRG_STRC", "LRN_NDIM_SUBSPACE"],
    effectsgroups=[
        EffectsGroup(
            scope=HasSpecies() & OwnedBy(empire=Source.Owner),
            accountinglabel="CON_TECH_ACCOUNTING_LABEL",
            priority=TARGET_POPULATION_AFTER_SCALING_PRIORITY,
            effects=[
                SetTargetConstruction(value=Value + 10),
                SetTargetPopulation(value=Value + 2 * Target.HabitableSize),
            ],
        )
    ],
    graphic="icons/tech/n-dimensional_structures.png",
)



from expansion_plans.expansion_plans_interface import (
    colonies_targeted,
    get_colonisable_outpost_ids,
    get_colonisable_planet_ids,
    outposts_targeted,
    update_colonisable_outpost_ids,
    update_colonisable_planet_ids,
)



from abc import ABC
from enum import Enum
from typing import Callable, TypeVar, Union

from common.statistic_interface._dump_value import DumpDict, DumpInt, DumpMultiple
from common.statistic_interface._serizlizer import to_float, to_int, to_str

_ScalarValue = Union[str, int, float]

StatValue = TypeVar("StatValue", dict[str, _ScalarValue], _ScalarValue)


class Serializer(ABC):
    def __init__(
        self,
        *,
        serializer: Callable[[StatValue], str],
        deserializer: Callable[[str], StatValue],
    ):
        self._deserializer = deserializer
        self._serializer = serializer

    def deserialize(self, value: str) -> StatValue:
        return self._deserializer(value)

    def serialize(self, value: StatValue) -> str:
        return self._serializer(value)


class StatKey(Enum):
    PolicyAdoption = DumpMultiple("PolicyAdoption", to_str)
    PolicyDeAdoption = DumpMultiple("PolicyDeAdoption", to_str)
    EmpireColors = DumpDict(
        "EmpireColors",
        {
            "R": to_int,
            "G": to_int,
            "B": to_int,
            "A": to_int,
        },
    )
    EmpireID = DumpDict("EmpireID", {"empire_id": to_int, "name": to_str, "turn": to_int})
    CapitalID = DumpDict(
        "CapitalID",
        {
            "capital_planet_id": to_int,
            "capital_planet_name": to_str,
            "capital_species": to_str,
        },
    )
    Output = DumpDict("CurrentOutput", {"turn": to_int, "RP": to_float, "PP": to_float})
    SHIP_CONT = DumpInt("ShipCount")

    def is_multi(self):
        return self.value.multi

    @classmethod
    def get_by_value_name(cls, name: str) -> "StatKey":
        for item in cls:
            if item.value.name == name:
                return item
        raise ValueError(f"Cannot find value with {name}")


LOG_PREFIX = "##"



from focs._effects import (
    EffectsGroup,
    Focus,
    HasSpecies,
    OwnedBy,
    Planet,
    SetMaxStockpile,
    Source,
    Target,
    Value,
)
from focs._tech import *
from macros.base_prod import STOCKPILE_PER_POP, TECH_COST_MULTIPLIER

Tech(
    name="PRO_INTERSTELLAR_ENTANGLEMENT_FACTORY",
    description="PRO_INTERSTELLAR_ENTANGLEMENT_FACTORY_DESC",
    short_description="IMPERIAL_STOCKPILE_SHORT_DESC",
    category="PRODUCTION_CATEGORY",
    researchcost=160 * TECH_COST_MULTIPLIER,
    researchturns=5,
    tags=["PEDIA_PRODUCTION_CATEGORY"],
    prerequisites=["PRO_GENERIC_SUPPLIES", "PRO_ADAPTIVE_AUTOMATION", "LRN_GRAVITONICS"],
    unlock=Item(type=UnlockBuilding, name="BLD_STOCKPILING_CENTER"),
    effectsgroups=[
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & HasSpecies(),
            accountinglabel="INTERSTELLAR_ENTANGLEMENT_FACTORY_POPULATION_BONUS_LABEL",
            effects=SetMaxStockpile(value=Value + 1.0 * Target.Population * STOCKPILE_PER_POP),
        ),
        EffectsGroup(
            scope=Planet() & OwnedBy(empire=Source.Owner) & Focus(type=["FOCUS_STOCKPILE"]),
            accountinglabel="INTERSTELLAR_ENTANGLEMENT_FACTORY_FIXED_BONUS_LABEL",
            effects=SetMaxStockpile(value=Value + 6),
        ),
    ],
)



from time import time


class Timer:
    def __init__(self, timer_name):
        """
        Creates timer. Timer name is name that will be in logs.
        """
        self.timer_name = timer_name
        self.start_time = None
        self.end_time = None
        self.section_name = None
        self.timers = []

    def stop(self, section_name=""):
        """
        Stop timer if running. Specify section_name if want to override its name.
        """
        if self.start_time:
            self.end_time = time()
            section_name = section_name or self.section_name
            self.timers.append((section_name, (self.end_time - self.start_time) * 1000.0))
        self.start_time = None
        self.section_name = None

    def start(self, section_name):
        """
        Stop prev timer if present and starts new.
        """
        self.stop()
        self.section_name = section_name
        self.start_time = time()

    def _print_timer_table(self, time_table):
        """
        Print a header and a footer and fill it with the timing results from time_table.
        """
        # TODO: Convert to use table code from here /python/common/print_utils.py#L116

        if not time_table:
            print("NOT Time Table")
            return
        max_header = max(len(x[0]) for x in time_table)
        line_max_size = max_header + 14
        print()
        print("Timing for %s:" % self.timer_name)
        print("=" * line_max_size)

        for name, val in time_table:
            print("%-*s %8d msec" % (max_header, name, val))

        print("-" * line_max_size)
        print(("Total: %8d msec" % sum(x[1] for x in time_table)).rjust(line_max_size))

    def stop_print_and_clear(self):
        """
        Stop timer, output result and clear state.
        """
        self.stop()
        self._print_timer_table(self.timers)
        self._clear_data()

    def _clear_data(self):
        """
        Clear timer data.
        """
        self.timers = []



import freeOrionAIInterface as fo
from collections import OrderedDict
from collections import OrderedDict as odict
from collections.abc import Iterable
from itertools import chain
from logging import debug

from aistate_interface import get_aistate
from common.fo_typing import PlanetId, SpeciesName
from EnumsAI import MissionType
from expansion_plans import expansion_plans_interface
from FleetUtilsAI import get_targeted_planet_ids


def initialise_expansion_plans():
    expansion_plans_interface._the_planner = ExpansionPlanner()


class ExpansionPlanner:
    @staticmethod
    def colonies_targeted() -> frozenset[PlanetId]:
        colony_ship_pids = get_targeted_planet_ids(fo.getUniverse().planetIDs, MissionType.COLONISATION)
        colony_building_pids = [e.locationID for e in fo.getEmpire().productionQueue if e.name.startswith("BLD_COL_")]
        debug(f"colonies_targeted: colony_ship_pids={colony_ship_pids}, colony_building_pids={colony_building_pids}")
        return frozenset(chain(colony_ship_pids, colony_building_pids))

    @staticmethod
    def outposts_targeted() -> frozenset[PlanetId]:
        outpost_ship_pids = get_targeted_planet_ids(fo.getUniverse().planetIDs, MissionType.OUTPOST)
        outpost_base_pids = get_targeted_planet_ids(fo.getUniverse().planetIDs, MissionType.ORBITAL_OUTPOST)
        debug(f"outpost_targeted: outpost_ship_pids={outpost_ship_pids}, outpost_base_pids={outpost_base_pids}")
        return frozenset(chain(outpost_ship_pids, outpost_base_pids))

    @staticmethod
    def get_colonisable_planet_ids(include_targeted: bool = False) -> OrderedDict[PlanetId, tuple[float, SpeciesName]]:
        if include_targeted:
            return get_aistate().colonisablePlanetIDs
        else:
            return odict(
                (pid, values)
                for (pid, values) in get_aistate().colonisablePlanetIDs.items()
                if pid not in expansion_plans_interface.colonies_targeted()  # interface function is cached
            )

    @staticmethod
    def get_colonisable_outpost_ids(include_targeted: bool = False) -> OrderedDict[PlanetId, tuple[float, SpeciesName]]:
        if include_targeted:
            return get_aistate().colonisableOutpostIDs
        else:
            return odict(
                (pid, values)
                for (pid, values) in get_aistate().colonisableOutpostIDs.items()
                if pid not in expansion_plans_interface.outposts_targeted()  # interface function is cached
            )

    @staticmethod
    def update_colonisable_planet_ids(new_list: Iterable) -> None:
        get_aistate().colonisablePlanetIDs.clear()
        get_aistate().colonisablePlanetIDs.update(new_list)

    @staticmethod
    def update_colonisable_outpost_ids(new_list: Iterable) -> None:
        get_aistate().colonisableOutpostIDs.clear()
        get_aistate().colonisableOutpostIDs.update(new_list)



# Increase the fuel capacity of ships with FU_BASIC_TANK parts.
# Adds the given increase amount times the number of FU_BASIC_TANK parts in the design to the max fuel meter.
# @1@ Accounting label to present this effect with
# @2@ Fuel capacity increase per fuel part
from focs._effects import (
    DesignHasPart,
    EffectsGroup,
    NamedReal,
    OwnedBy,
    PartsInShipDesign,
    SetMaxFuel,
    Ship,
    Source,
    Target,
    Value,
)


def PART_UPGRADE_MAXFUEL_EFFECTS(effect_name: str, value: float):
    return EffectsGroup(
        scope=Ship & OwnedBy(empire=Source.Owner) & DesignHasPart(name="FU_BASIC_TANK"),
        accountinglabel=effect_name,
        effects=[
            SetMaxFuel(
                value=Value
                + NamedReal(name=effect_name + "_MULT", value=value)
                * PartsInShipDesign(name="FU_BASIC_TANK", design=Target.DesignID)
            )
        ],
    )



from focs._effects import (
    EffectsGroup,
    IsSource,
    SetEmpireMeter,
    Source,
    Value,
)
from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="LRN_MIND_VOID",
    description="LRN_MIND_VOID_DESC",
    short_description="POLICY_AND_SLOT_UNLOCK_SHORT_DESC",
    category="LEARNING_CATEGORY",
    researchcost=120 * TECH_COST_MULTIPLIER,
    researchturns=10,
    tags=["PEDIA_LEARNING_CATEGORY", "THEORY"],
    prerequisites=["LRN_XENOARCH"],
    unlock=Item(type=UnlockPolicy, name="PLC_DESIGN_SIMPLICITY"),
    effectsgroups=[
        EffectsGroup(
            scope=IsSource,
            effects=SetEmpireMeter(empire=Source.Owner, meter="SOCIAL_CATEGORY_NUM_POLICY_SLOTS", value=Value + 1),
        )
    ],
    graphic="icons/tech/mind_of_the_void.png",
)



from focs._tech import *
from macros.base_prod import TECH_COST_MULTIPLIER

Tech(
    name="SHP_SOLAR_CONT",
    description="SHP_SOLAR_CONT_DESC",
    short_description="SHIP_HULL_UNLOCK_SHORT_DESC",
    category="SHIP_HULLS_CATEGORY",
    researchcost=1120 * TECH_COST_MULTIPLIER,
    researchturns=7,
    tags=["PEDIA_ENERGY_HULL_TECHS"],
    prerequisites=["SHP_QUANT_ENRG_MAG", "SHP_ENRG_BOUND_MAN", "PRO_ZERO_GEN"],
    unlock=[
        Item(type=UnlockShipHull, name="SH_SOLAR")
        # Item type = Building name = "BLD_SHIPYARD_ENRG_SOLAR"
    ],
    graphic="icons/ship_hulls/solar_hull_small.png",
)
